
# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
import re
from copy import deepcopy
from typing import Union, Dict, Iterator

from pyparsing import Word, ParseException, nums, Literal, LineEnd, Suppress, alphanums
from networkx import MultiDiGraph, subgraph, find_cycle, NetworkXNoCycle, DiGraph
from pandas import DataFrame

from pyrates import PyRatesException
from pyrates.ir.node import NodeIR
from pyrates.ir.edge import EdgeIR
from pyrates.ir.abc import AbstractBaseIR

__author__ = "Daniel Rose"
__status__ = "Development"


class CircuitIR(AbstractBaseIR):
    """Custom graph data structure that represents a backend of nodes and edges with associated equations
    and variables."""

    # _node_label_grammar = Word(alphanums+"_") + Suppress(".") + Word(nums)

    def __init__(self, label: str = "circuit", circuits: dict = None, nodes: Dict[str, NodeIR] = None,
                 edges: list = None, template: str = None):
        """
        Parameters:
        -----------
        label
            String label, could be used as fallback when subcircuiting this circuit. Currently not used, though.
        circuits
            Dictionary of sub-circuits to be added. Keys are string labels for circuits that serve as namespaces for the
            subcircuits. Items must be `CircuitIR` instances.
        nodes
            Dictionary of nodes of form {node_label: `NodeIR` instance}.
        edges
            List of tuples (source:str, target:str, edge_dict). `edge_dict` should contain the key "edge_ir" with an
            `EdgeIR` instance as item and optionally entries for "weight" and "delay". `source` and `target` should be
            formatted as "node/op/var" (with optionally prepended circuits).
        template
            optional string reference to path to template that this circuit was loaded from. Leave empty, if no template
            was used.
        """

        super().__init__(template)
        self.label = label
        self.label_counter = {}
        self.label_map = {}

        self.graph = MultiDiGraph()
        self.sub_circuits = set()

        if circuits:
            for key, temp in circuits.items():
                self.add_circuit(key, temp)

        if nodes:
            self.add_nodes_from(nodes)

        if edges:
            self.add_edges_from(edges)

    def add_nodes_from(self, nodes: Dict[str, NodeIR], **attr):
        """ Add multiple nodes to circuit. Allows networkx-style adding if from_templates is set to False.

        Parameters
        ----------
        nodes
            Dictionary with node label as key. The item is a NodeIR instance. Note that the item type is not tested
            here, but passing anything that does not behave like a `NodeIR` may cause problems later.
        attr
            additional keyword attributes that can be added to the node data. (default `networkx` syntax.)
        """

        # get unique labels for nodes
        for label in nodes:
            self.label_map[label] = self._get_unique_label(label)

        # rename node keys
        # assign NodeIR instances as "node" keys in a separate dictionary, because networkx saves node attributes into
        # a dictionary
        # reformat dictionary to tuple/generator, since networkx does not parse dictionary correctly in add_nodes_from
        nodes = ((self.label_map[key], {"node": node}) for key, node in nodes.items())
        self.graph.add_nodes_from(nodes, **attr)

    def add_node(self, label: str, node: NodeIR, **attr):
        """Add single node

        Parameters
        ----------
        label
            String to identify node by. Is tested for uniqueness internally, and renamed if necessary. Renamed labels
            are stored in the `CircuitIR` instance attribute `label_map`.
        node
            Instance of `NodeIR`. Will be added with the key "node" to the node dictionary.
        attr
            Additional attributes (keyword arguments) that can be added to the node data. (Default `networkx` syntax.)
        """

        new_label = self._get_unique_label(label)

        self.graph.add_node(new_label, node=node, **attr)

        self.label_map[label] = new_label

    def add_edges_from(self, edges, **attr):
        """ Add multiple edges. This method explicitly assumes, that edges are given in edge_templates instead of
        existing instances of `EdgeIR`.

        Parameters
        ----------
        edges
            List of edges, each of shape [source/op/var, target/op/var, edge_dict]. The edge_dict must contain the
            keys "edge_ir", and optionally "weight" and "delay".
        attr
            Additional attributes (keyword arguments) that can be added to the edge data. (Default `networkx` syntax.)


        Returns
        -------
        """

        edge_list = []
        for (source, target, edge_dict) in edges:
            # get weight
            weight = edge_dict.get("weight", 1.)
            # get delay
            delay = edge_dict.get("delay", None)

            # get edge_ir or (if not included) default to an empty edge
            edge_ir = edge_dict.get("edge_ir", EdgeIR())

            if "target_var" in edge_dict:
                target_var = edge_dict["target_var"]
                target = f"{target}/{target_var}"

            if "source_var" in edge_dict:
                source_var = edge_dict["source_var"]
                source = f"{source}/{source_var}"

            # test, if variables at source and target exist and reference them properly
            source, target = self._validate_separate_key_path(source, target)

            edge_list.append((source[0], target[0],  # edge_unique_key,
                              {"edge_ir": edge_ir,
                               "weight": weight,
                               "delay": delay,
                               "source_var": "/".join(source[-2:]),
                               "target_var": "/".join(target[-2:])
                               }))
        self.graph.add_edges_from(edge_list, **attr)

    def add_edge(self, source: str, target: str, edge_ir: EdgeIR, weight: float = 1., delay: float = None,
                 identify_relations=True,
                 **data):
        """
        Parameters
        ----------
        source
        target
        edge_ir
        weight
        delay
        data
            If no template is given, `data` is assumed to conform to the format that is needed to add an edge. I.e.,
            `data` needs to contain fields for `weight`, `delay`, `edge_ir`, `source_var`, `target_var`.
        identify_relations

        Returns
        -------

        """

        source_var = ""
        target_var = ""
        if identify_relations:
            # test, if variables at source and target exist and reference them properly
            source, target = self._validate_separate_key_path(source, target)
        else:
            # assume that source and target strings are already consistent. This should be the case,
            # if the given strings were coming from existing circuits (instances of `CircuitIR`)
            # or in general, if operators are not renamed.
            for path in (source, target):
                if path not in self:
                    raise PyRatesException(f"Failed to add edge, because referenced node `{path}` does not exist in "
                                           f"network graph. Edges can only be added to existing nodes.")

            source_var = data.pop("source_var", "")
            target_var = data.pop("target_var", "")
            source = source.split("/")
            target = target.split("/")

        # temporary workaround to make sure source/target variable/operator and nodes are defined properly
        if source_var:
            source_node = "/".join(source)
        else:
            source_node = "/".join(source[:-2])
            source_var = "/".join(source[-2:])

        if target_var:
            target_node = "/".join(target)
        else:
            target_node = "/".join(target[:-2])
            target_var = "/".join(target[-2:])

        attr_dict = dict(edge_ir=edge_ir,
                         weight=weight,
                         delay=delay,
                         source_var=source_var,
                         target_var=target_var,
                         **data)

        self.graph.add_edge(source_node, target_node, **attr_dict)

    def _get_unique_label(self, label: str) -> str:
        """Tests, if a given node `label` already exists in the circuit and renames it uniquely, if necessary.
        Uniqueness is generally ensure by appending a counter of the form ".0" . If the given label already has a
        counter, it will be detected ond potentially altered if uniqueness requires it. The resulting (new) label is
        returned.

        Parameters
        ----------
        label

        Returns
        -------
        unique_label
        """
        # test, if label already has a counter and separate it, if necessary
        # could use pyparsing instead of regex, but actually seems more robust and simple enough in this case
        match = re.match("(.+)[.]([\d]+$)", label)
        if match:
            # see if label already exists, just continue if it doesn't
            if label in self:
                # create new label
                # separate base label and counter
                label, counter = match.groups()
                counter = int(counter)
                if label not in self.label_counter:
                    self.label_counter[label] = counter + 1
                elif counter > self.label_counter[label]:
                    self.label_counter[label] = counter + 1
                else:
                    self.label_counter[label] += 1
                # set label
                unique_label = f"{label}.{self.label_counter[label]}"
            else:
                # pass original label
                unique_label = label

        else:
            # define counter
            if label in self.label_counter:
                self.label_counter[label] += 1
            else:
                self.label_counter[label] = 0

            # set label
            unique_label = f"{label}.{self.label_counter[label]}"

        return unique_label

    def _validate_separate_key_path(self, *paths: str):

        for key in paths:

            # (circuits), node, operator and variable specifiers

            *node, op, var = key.split("/")

            node = "/".join(node)

            # re-reference node labels, if necessary
            # this syntax yields "node" back as default if it is not in label_map
            node = self.label_map.get(node, node)
            # re_reference operator labels, if necessary
            op = self._validate_rename_op_label(self[node], op)
            # ignore circuits for now
            # note: current implementation assumes, that this method is only called, if an edge is added
            path = "/".join((node, op, var))
            # check if path is valid
            if path not in self:
                raise PyRatesException(f"Could not find object with key path `{path}`.")

            separated = (node, op, var)
            yield separated

    @staticmethod
    def _validate_rename_op_label(node: NodeIR, op_label: str) -> str:
        """
        References to operators in source/target references may mismatch actual operator labels, due to internal renaming
        that can not be accounted for in the YAML templates. This method looks for the first instance of an operator in
        a specific node, assuming it exists at all. Additionally, this assumes, that only one variation of an operator
        template (of same name) can exist in any single node.
        Note: The latter assumption becomes invalid, if edge operators are moved to nodes, because in that case multiple
        variations of the same operator can exist in one node.

        Parameters
        ----------
        node
        op_label

        Returns
        -------
        new_op_label

        """
        # variant 1: label already has a counter --> might actually already exist.

        if "." in op_label:
            try:
                _ = node[op_label]
            except KeyError:
                op_label, *_ = op_label.split(".")
            else:
                return op_label

        # build grammar for pyparsing
        grammar = Literal(op_label) + "." + Word(nums)
        found = 0  # count how many matching operators were found
        for op_key in node:
            try:
                grammar.parseString(op_key, parseAll=True)
            except ParseException:
                continue
            else:
                op_label = op_key
                found += 1

        if found == 1:
            return op_label
        elif not found:
            raise PyRatesException(f"Could not identify operator with base name '{op_label}' "
                                   f"in node `{node}`.")
        else:
            raise PyRatesException(f"Unable to uniquely identify operator key. "
                                   f"Found multiple occurrences for operator with base name '{op_label}'.")

    def getitem_from_iterator(self, key: str, key_iter: Iterator[str]):

        if key in self.sub_circuits:
            item = SubCircuitView(self, key)
        else:
            item = self.graph.nodes[key]["node"]

        return item

    @property
    def nodes(self):
        """Shortcut to self.graph.nodes. See documentation of `networkx.MultiDiGraph.nodes`."""
        return self.graph.nodes

    @property
    def edges(self):
        """Shortcut to self.graph.edges. See documentation of `networkx.MultiDiGraph.edges`."""
        return self.graph.edges

    @classmethod
    def from_circuits(cls, label: str, circuits: dict, connectivity: Union[list, tuple, DataFrame] = None):
        """Circuit creation method that takes multiple circuits (templates or instances of `CircuitIR`) as inputs to
        create one larger circuit out of these. With additional `connectivity` information, these circuit can directly
        be interlinked.

        Parameters
        ----------
        label
            Name of new circuit. Should not collide with any circuit label given in `circuits`.
        circuits
            Dictionary with unique circuit labels as keys and circuits as items. Circuits may either be instances of
            `CircuitTemplate` or `CircuitIR`. Alternatively, a circuit template may also be given via a sub-dictionary
            with keys `template` and `values`, where `values` is a dictionary of variable value updates for the given
            template.
        connectivity
            Optional `list`, `tuple` or `pandas.DataFrame' with connectivity information to create edges between the
            given circuits. If `list` or `tuple`, then each item must be formatted the same way as `edges` in
            `add_edges_from`: ('circuit/source_node/op/var', 'circuit/target_node/op/var', edge_template, variables).
            If given as a `DataFrame`, keys (indices and column names) must refer to sources and targets, respectively,
            as column name/index (string of form 'circuit/node/op/var') and items may then be edge templates and
            associated variables.
            Empty cells in the DataFrame should be filled with something 'falsy' (as in evaluates to `False` in Python).

        Returns
        -------
        circuit
            instance of `CircuitIR`
        """
        # ToDo: Rewrite doc to account for assumption, that only CircuitIR instances are allowed

        circuit = cls(label, nodes={}, edges=[])
        for name, circ in circuits.items():
            circuit.add_circuit(name, circ)

        if connectivity is not None:
            if isinstance(connectivity, list) or isinstance(connectivity, tuple):
                circuit.add_edges_from(connectivity)
            else:
                try:
                    if isinstance(connectivity, dict):
                        key, conn_info = connectivity.popitem()
                        for target, row in conn_info.iterrows():
                            for source, content in row.iteritems():
                                snode, tnode = source.split('/')[:-2], target.split('/')[:-2]
                                svar, tvar = source.split('/')[-2:], target.split('/')[-2:]
                                snode, tnode = "/".join(snode), "/".join(tnode)
                                svar, tvar = "/".join(svar), "/".join(tvar)
                                content = {key: content} if content else {}
                                for key_tmp, conn_info_tmp in connectivity.items():
                                    content_tmp = conn_info_tmp.loc[target, source]
                                    if content_tmp:
                                        content.update({key_tmp: content_tmp})
                                content.update({'source_var': svar, 'target_var': tvar})
                                if 'weight' in content and content['weight']:
                                    circuit.add_edge(snode, tnode, edge_ir=None, identify_relations=False,
                                                     **content)
                    else:
                        for target, row in connectivity.iterrows():
                            for source, content in row.iteritems():
                                if content:  # assumes, empty entries evaluate to `False`
                                    snode, tnode = source.split('/')[:-2], target.split('/')[:-2]
                                    svar, tvar = source.split('/')[-2:], target.split('/')[-2:]
                                    snode, tnode = "/".join(snode), "/".join(tnode)
                                    svar, tvar = "/".join(svar), "/".join(tvar)
                                    if "float" in str(type(content)):
                                        content = {'weight': content, 'delay': None}
                                    content.update({'source_var': svar, 'target_var': tvar})
                                    circuit.add_edge(snode, tnode, edge_ir=None, identify_relations=False, **content)
                except AttributeError:
                    raise TypeError(f"Invalid data type of variable `connectivity` (type: {type(connectivity)}).")

        return circuit

    def add_circuit(self, label: str, circuit):
        """ Add a single circuit (with its own nodes and edges) to this circuit (like a subgraph in a graph).

        Parameters
        ----------
        label
            Assigned name of the circuit. If this name is already in use, the label will be renamed in the form
            `label.idx`.
        circuit
            Instance of `CircuitIR` or `CircuitTemplate` or a dictionary, where the key 'template' refers to a
            `CircuitTemplate` instance and 'values' refers to updates that should be applied to the template.
        Returns
        -------

        """
        # ToDo: disallow usage of templates here

        # parse data type of circuit
        if isinstance(circuit, dict):
            circuit = circuit["template"].apply(circuit["values"])  # type: CircuitIR
        else:
            try:
                # if it is a template, apply it
                circuit = circuit.apply()  # type: CircuitIR
            except AttributeError:
                # assume circuit already is a circuitIR or similarly structured construct
                pass

        # check if given circuit label already exists in this circuit
        if label in self.sub_circuits:
            raise PyRatesException(f"Circuit label {label} already exists in this circuit. Please specify a unique "
                                   f"circuit label.")
            # may change to a rule to rename circuits (like circuit.0, circuit.1, circuit.2...) with label map and
            # counter

        # add circuit nodes, node by node, appending circuit label to node name
        for name, data in circuit.nodes(data=True):
            self.add_node(f"{label}/{name}", **data)

        # add circuit reference to sub_circuits set. Needs to be done before adding edges
        self.sub_circuits.add(label)
        for sc in circuit.sub_circuits:
            self.sub_circuits.add(f"{label}/{sc}")

        # add sub circuit label map items to local label map
        for old, new in circuit.label_map.items():
            self.label_map[f"{label}/{old}"] = f"{label}/{new}"

        # add edges
        for source, target, data in circuit.edges(data=True):
            # source_var = data.pop("source_var")
            # target_var = data.pop("target_var")
            self.add_edge(f"{label}/{source}", f"{label}/{target}", identify_relations=False, **data)

    def move_edge_operators_to_nodes(self, copy_data=True):
        """Returns a new CircuitIR instance, where all operators that were previously in edges are moved to their
        respective target nodes."""
        # if copy_data:
        #    nodes = {key: deepcopy(data) for key, data in self.nodes(data=True)}
        # else:
        #    nodes = {key: data for key, data in self.nodes(data=True)}

        # this does not preserve additional node attributes
        # node_attrs = {}
        # for key, data in nodes.items():
        #    nodes[key] = data["node"]
        # node_attrs.update(**data)

        # node_attrs.pop("node")

        op_label_counter = {}

        # edges = []
        for source, target, data in self.edges(data=True):

            if "edge_ir" in data and data["edge_ir"] and data["edge_ir"].op_graph:
                source_var = data["source_var"]
                target_var = data["target_var"]
                weight = data["weight"]
                delay = data["delay"]

                edge_ir = data["edge_ir"]  # type: EdgeIR
                op_graph = edge_ir.op_graph
                if copy_data:
                    op_graph = deepcopy(op_graph)
                input_var = edge_ir.input
                output_var = edge_ir.output
                if len(op_graph) > 0:
                    target_var = self._move_ops_to_target(target, input_var, output_var, target_var, op_graph,
                                                          op_label_counter, self.nodes)
                    # side effect: changes op_label_counter and nodes dictionary

                data.update({'source_var': source_var, 'target_var': target_var, 'edge_ir': EdgeIR(), 'weight': weight,
                             'delay': delay})
            # edges.append((source, target, data))

        # circuit = CircuitIR()
        # circuit.sub_circuits = self.sub_circuits
        # circuit.add_nodes_from(nodes)
        # circuit.add_edges_from(edges)
        # return circuit

        return self

    @staticmethod
    def _move_ops_to_target(target, input_var, output_var, target_var, op_graph: DiGraph, op_label_counter, nodes):
        # check, if cycles are present in operator graph (which would be problematic
        try:
            find_cycle(op_graph)
        except NetworkXNoCycle:
            pass
        else:
            raise PyRatesException("Found cyclic operator graph. Cycles are not allowed for operators within one node.")

        target_node = nodes[target]["node"]
        if target not in op_label_counter:
            op_label_counter[target] = {}

        key_map = {}
        # first pass: get all labels right
        for key in op_graph.nodes:
            # rename operator key by appending another counter
            if key in op_label_counter[target]:
                # already renamed it previously --> increase counter
                op_label_counter[target][key] += 1
                key_map[key] = f"{key}.{op_label_counter[target][key]}"

            else:
                # not renamed previously --> initialize counter to 0
                op_label_counter[target][key] = 0
                key_map[key] = f"{key}.0"

        # second pass: add operators to target op graph and rename sources according to key_map
        for key, data in op_graph.nodes(data=True):
            op = data["operator"]
            variables = data["variables"]
            # go through all input variables and rename source operators
            for var, var_data in op.inputs.items():
                sources = []
                for source_op in var_data["sources"]:
                    sources.append(key_map[source_op])
                op.inputs[var]["sources"] = sources

            # add operator to target_node's operator graph
            # TODO: implement add_node method on OperatorGraph class
            target_node.op_graph.add_node(key_map[key], operator=op, variables=variables)

        # add edges that previously existed
        for source_op, target_op in op_graph.edges:
            # rename edge keys accordingly
            source_op = key_map[source_op]
            target_op = key_map[target_op]

            target.op_graph.add_edge(source_op, target_op)

        # add new edges based on output and target of edge
        target_op, target_var = target_var.split("/")
        output_op, output_var = output_var.split("/")
        output_op = key_map[output_op]

        target_node[target_op].inputs[target_var]["sources"].append(output_op)

        # make sure the changes are saved (might not be necessary)
        # nodes[target] = target_node

        # reassign target variable of edge
        target_op, target_var = input_var.split("/")
        return f"{key_map[target_op]}/{target_var}"


class SubCircuitView(AbstractBaseIR):
    """View on a subgraph of a circuit. In order to keep memory footprint and computational cost low, the original (top
    lvl) circuit is referenced locally as 'top_level_circuit' and all subgraph-related information is computed only
    when needed."""

    def __init__(self, top_level_circuit: CircuitIR, subgraph_key: str):

        super().__init__()
        self.top_level_circuit = top_level_circuit
        self.subgraph_key = subgraph_key

    def getitem_from_iterator(self, key: str, key_iter: Iterator[str]):

        key = f"{self.subgraph_key}/{key}"

        if key in self.top_level_circuit.sub_circuits:
            return SubCircuitView(self.top_level_circuit, key)
        else:
            return self.top_level_circuit.nodes[key]["node"]

    @property
    def induced_graph(self):
        """Return the subgraph specified by `subgraph_key`."""

        nodes = (node for node in self.top_level_circuit.nodes if node.startswith(self.subgraph_key))
        return subgraph(self.top_level_circuit.graph, nodes)

    def __str__(self):

        return f"{self.__class__.__name__} on '{self.subgraph_key}' in {self.top_level_circuit}"




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
import re
from copy import deepcopy
from typing import List, Union, Iterator

from pyrates import PyRatesException
from pyrates.frontend.utility import deep_freeze
from pyrates.ir.abc import AbstractBaseIR

__author__ = "Daniel Rose"
__status__ = "Development"


class OperatorIR(AbstractBaseIR):

    def __init__(self, equations: List[str], inputs: list, output: str, template: str=None):

        super().__init__(template)
        self.output = output
        self.inputs = inputs
        self.equations = equations

    # @staticmethod
    # def _reduce_ode_order(equations: str, variables):
    #     """Checks if a 2nd-order ODE is present and reduces it to two coupled first-order ODEs.
    #     Currently limited to special case of the form '(d/dt + a)^2 * x = b'.
    #
    #     Parameters
    #     ----------
    #     equations
    #         string of form 'a = b'
    #     """
    #
    #     # matches pattern of form `(d/dt + a)^2 * y` and extracts `a` and `y`
    #     match = re.match(r"\(\s*d\s*/\s*dt\s*[+-]\s*([\d]*[.]?[\d]*/?[a-zA-Z]\w*)\s*\)\s*\^2\s*\*\s*([a-zA-Z]\w*)",
    #                      equations)
    #
    #     if match:
    #         # assume the entire lhs was matched, fails if there is something remaining on the lhs
    #         lhs, rhs = equations.split("=")
    #         a, var = match.groups()  # returns coefficient `a` and variable `y`
    #         eq1 = f"d/dt * {var} = {var}_t"
    #         eq2 = f"d/dt * {var}_t = {rhs} - ({a})^2 * {var} - 2. * {a} * {var}_t"
    #
    #         variables[f"{var}_t"] = {"dtype": "float32",
    #                                  "description": "integration variable",
    #                                  "vtype": "state_var",
    #                                  "value": variables[var]['value'] if variables[var]['value'] else 0.}
    #
    #         return eq1, eq2, variables
    #     else:
    #         return equations, variables

    def copy(self):

        return self.__class__(deepcopy(self.equations), deepcopy(self.inputs), deepcopy(self.output))

    def getitem_from_iterator(self, key: str, key_iter: Iterator[str]):
        """
        Checks if a variable named by key exists in an equations.
        Parameters
        ----------
        key
        key_iter

        Returns
        -------
        key
        """

        for equation in self.equations:
            if key in equation:
                return key
        else:
            raise KeyError(f"Variable `{key}` not found in equations {self.equations}")



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
from typing import Iterator

from pyrates.ir.abc import AbstractBaseIR
from pyrates.ir.operator_graph import OperatorGraph

__author__ = "Daniel Rose"
__status__ = "Development"


class NodeIR(AbstractBaseIR):

    def __init__(self, operators: dict=None, template: str=None):

        super().__init__(template)
        self.op_graph = OperatorGraph(operators)

    def getitem_from_iterator(self, key: str, key_iter: Iterator[str]):
        """Alias for self.op_graph.getitem_from_iterator"""

        return self.op_graph.getitem_from_iterator(key, key_iter)

    def __iter__(self):
        """Return an iterator containing all operator labels in the operator graph."""
        return iter(self.op_graph)

    @property
    def operators(self):
        return self.op_graph.operators



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

# external imports
import numpy as np
import pandas as pd
from typing import Optional, Union

# system imports
import os
import random
from itertools import cycle

from pyrates.utility.grid_search import grid_search, ClusterGridSearch, linearize_grid

# meta infos
__author__ = "Christoph Salomon, Richard Gast"
__status__ = "development"


class GeneticAlgorithmTemplate:
    def __init__(self):

        # Initialize storage variables
        self.initial_gene_pool = {}
        self.num_genes = 0
        self.sigma_adapt = 0
        self.gene_names = []
        self.pop = pd.DataFrame()
        self.pop_size = 0
        self.candidate = pd.DataFrame()
        self.winner = pd.DataFrame()
        self.current_max_fitness = 0
        self.drop_count = 0

    def run(self, initial_gene_pool: dict, target: list, max_iter: int, min_fit: Optional[float] = 0.,
            n_winners: Optional[int] = 1, n_parent_pairs: Optional[int] = 10, n_new: Optional[int] = 0,
            sigma_adapt: Optional[float] = 0., max_stagnation_steps: Optional[int] = 0,
            stagnation_decimals: Optional[int] = 8, max_stagnation_drops: Optional[Union[int, float]] = np.Inf,
            enforce_max_iter: Optional[bool] = False, candidate_save: Optional[str] = "", drop_save: Optional[str] = "",
            gene_sampling_func=np.linspace, **kwargs):
        """Run a genetic algorithm to fit genes of a population in respect to a given target vector

        Parameters
        ----------
        initial_gene_pool
        target
            Target values that are used to determine the fitness of a population member
        max_iter
            Maximum number of iterations (generations) before the computation is terminated
        min_fit
            Minimum fitness. If set, computation will stop after the fitness of one population member exceeds this value
        n_winners
            Number of strongest members of a population, that will be members of this population's offspring
        n_parent_pairs
            Number of parent pairs of a population that will produce a child which will be member of this population's
            offspring (crossover)
        n_new
            Number of members in a population's offspring, that will be created from the initial gene pool
        sigma_adapt
            Ratio by which the mean derivation of the gene probability distributions will change during a mutation
        max_stagnation_steps
            Maximum number of iterations with no change in the fitness, before the strongest member of a population
            will be discarded and not be part of this population's offspring anymore
        stagnation_decimals
            Decimal precision that will be used to detect changes in the fitness of successive populations
        max_stagnation_drops
            If True, computation will stop when the maximum stagnation is reached
        enforce_max_iter
            If True, all iterations will performed, even if another convergence criterion is reached before
        candidate_save
            If set, the strongest member of a population will be saved to an hdf5 file, before the population is updated
        drop_save
            If set, all members that are droped from a population due to stagnation or other criteria will be saved to
            this folder in hdf5 format
        gene_sampling_func
        kwargs


        Returns
        -------
        pandas.DataFrame containing the overall fittest member of all computed populations

        """

        self.initial_gene_pool = initial_gene_pool
        self.num_genes = len(initial_gene_pool)
        self.sigma_adapt = sigma_adapt
        self.drop_count = 0

        # Create starting population
        ############################
        self.__create_pop(sampling_func=gene_sampling_func)
        self.pop_size = self.pop.shape[0]

        if n_winners + n_parent_pairs + n_new > self.pop_size:
            print('WARNING: Sum of winners, parents and new members exceeds the population size. Returning')
            return
        # Start genetic algorithm
        #########################
        print("***STARTING GENETIC ALGORITHM***")
        iter_count = 0
        stagnation_count = 0
        while iter_count < max_iter:
            print(f'ITERATION: {iter_count}')

            # Evaluate fitness of current population
            ########################################
            self.eval_fitness(target, **kwargs)
            new_candidate = self.pop.nlargest(1, "fitness")
            self.current_max_fitness = float(new_candidate.loc[:, "fitness"])
            print(f'Currently fittest genes:')
            self.plot_genes(new_candidate)
            target_tmp = [np.round(tar[0], decimals=2) for tar in target]
            print(f'Target: {target_tmp}')

            # Check for fitness stagnation
            ##############################
            if max_stagnation_steps > 0:
                if not self.candidate.empty:
                    old_fitness = np.round(float(self.candidate['fitness']), decimals=stagnation_decimals)
                    new_fitness = np.round(self.current_max_fitness, decimals=stagnation_decimals)
                    # Check for change in fitness
                    if new_fitness == old_fitness:
                        stagnation_count += 1
                        # Check if stagnation occured
                        if stagnation_count > max_stagnation_steps:
                            print("Maximum fitness stagnation reached!")
                            # Check if maximum number of population drops is reached
                            if not (self.drop_count == max_stagnation_drops) or enforce_max_iter:
                                print("Dropping fittest from population!")
                                if drop_save:
                                    os.makedirs(drop_save, exist_ok=True)
                                    new_candidate.to_hdf(f'{drop_save}/PopulationDrop_{self.drop_count}.h5', key='data')
                                self.drop_count += 1
                                self.pop = self.pop.drop(new_candidate.index)
                            else:
                                print("Returning fittest member!")
                                return new_candidate
                    else:
                        # Reset stagnation counter
                        stagnation_count = 0

            # Update candidate and save if necessary
            ########################################
            self.candidate = new_candidate
            if candidate_save:
                self.candidate.to_hdf(candidate_save, key='data')

            # Update current winning genes
            #############################
            if self.winner.empty:
                self.winner = self.candidate
            elif float(self.candidate['fitness']) > float(self.winner['fitness']):
                self.winner = self.candidate

            # Evaluate minimum fitness conversion criteria
            ##############################################
            if 0 < min_fit < self.current_max_fitness:
                print("Minimum fitness criterion reached!")
                if enforce_max_iter:
                    print("Dropping fittest from population!")
                    if drop_save:
                        new_candidate.to_hdf(f'{drop_save}/PopulationDrop_{self.drop_count}.h5', key='data')
                    self.drop_count += 1
                    self.pop = self.pop.drop(self.candidate.index)
                else:
                    return self.candidate

            # Drop additional information
            #############################
            try:
                self.pop = self.pop.drop("holgado", axis=1)
            except KeyError:
                pass

            # Create offspring from current population
            ##########################################
            self.__create_offspring(n_parent_pairs=n_parent_pairs, n_new=n_new, n_winners=n_winners)
            iter_count += 1

        # End of iteration loop
        print("Maximum iterations reached")
        if float(self.winner['fitness']) < min_fit:
            print('Could not satisfy minimum fitness condition.')
        return self.winner

    def __create_offspring(self, n_winners, n_parent_pairs=0, n_new=0):
        """Create a new offspring of the current population

        Offspring contains:
        - n_winners strongest members of the current population (winners)
        - n_parent_pairs Children of current parent pairings (crossover)
        - N Mutations of winners and children (mutations)
        - n_new Fresh members based on the initial gene pool (new)

        The number of mutations is chosen dynamically to resize the offspring to the size of the current population
        with N = population_size - n_winners - n_parent_pairs - n_new
        """
        print('Updating population')

        # Create new offspring
        ######################
        offspring = []
        new_sigs = []
        n_mutations = self.pop_size - (n_parent_pairs + n_new + n_winners)

        # 1. Add n_winners strongest members
        ####################################
        winners = self.__select_winners(n_winners=n_winners)
        for w in winners:
            offspring.append(w[0])
            new_sigs.append(w[1])

        # 2. Add children of n_parents parent pairs
        ###########################################
        parent_pairs = self.__create_parent_pairs(n_parent_pairs=n_parent_pairs)
        childs = []
        try:
            childs = self.__crossover(parent_pairs)
            for c in childs:
                offspring.append(c[0])
                new_sigs.append(c[1])
        except ValueError:
            # Each failed child will be replaced by a mutation
            n_mutations += (n_parent_pairs - len(childs))

        # 3. Add mutations
        ##################
        parent_pool = cycle(zip(offspring, new_sigs))
        for mut in range(n_mutations):
            parent = next(parent_pool)
            mutation = self.__mutate(parent)
            offspring.append(mutation[0])
            new_sigs.append(mutation[1])

        # 4. Add n_new fresh members from initial gene_pool
        ###################################################
        for n in range(n_new):
            new_member = self.__create_new_member()
            offspring.append(new_member[0])
            new_sigs.append(new_member[1])

        offspring = pd.DataFrame(offspring)
        offspring['fitness'] = 0.0
        offspring['sigma'] = new_sigs
        offspring.columns = self.pop.columns

        self.pop = offspring

    def __create_pop(self, sampling_func=np.linspace):
        """Create new population from the initial gene pool"""
        pop_grid = {}

        for param, value in self.initial_gene_pool.items():
            self.gene_names.append(param)
            if value['N'] == 1:
                pop_grid[param] = np.array([np.mean([value['min'], value['max']])])
            else:
                pop_grid[param] = sampling_func(value['min'], value['max'], value['N'])
            # pop_grid[param] = np.linspace(value['min'], value['max'], value['N'])
        self.pop = linearize_grid(pop_grid, permute=True)
        self.pop_size = self.pop.shape[0]

        sigmas = [self.initial_gene_pool[gene]['sigma'] for gene in self.initial_gene_pool.keys()]

        self.pop['fitness'] = 0.0
        self.pop['sigma'] = [sigmas for _ in range(self.pop_size)]

    def __select_winners(self, n_winners):
        """Returns the n_winners fittest members from the current population"""
        winners = []
        for idx in self.pop.nlargest(n_winners, 'fitness').index:
            winner_genes = self.pop.iloc[idx].drop(['fitness', 'sigma']).to_list()
            winner_sigma = self.pop.iloc[idx]['sigma']
            winners.append([winner_genes, winner_sigma])
        return winners

    def __mutate(self, parent):
        """Create mutation of a parent, based on a gaussian distribution for each gene"""
        mu_new = []
        sigma_new = []
        for i, (mu, sigma) in enumerate(zip(parent[0], parent[1])):
            mu_temp = mu + np.random.randn() * sigma
            while any([mu_temp < self.initial_gene_pool[self.gene_names[i]]['min'],
                       mu_temp > self.initial_gene_pool[self.gene_names[i]]['max']]):
                mu_temp = mu+np.random.randn()*sigma
            mu_new.append(mu_temp)
            # Adapt sigma (Beyer1995, p.5)
            xi = np.exp(self.sigma_adapt*np.random.randn())
            sigma_new.append(sigma*xi)
        return mu_new, sigma_new

    def __create_new_member(self):
        """Create a new population member from the initial gene pool"""
        genes = []
        sigma = []
        for i, (key, value) in enumerate(self.initial_gene_pool.items()):
            genes.append(random.uniform(value['min'], value['max']))
            sigma.append(value['sigma'])
        new_member = [genes, sigma]
        already_exists = (self.pop.drop(['fitness', 'sigma'], axis=1) == new_member[0]).all(1).any()
        if already_exists:
            new_member = self.__create_new_member()
        return new_member

    def __create_parent_pairs(self, n_parent_pairs):
        """Create n_parent_pairs parent combinations. The occurrence probability for each parent is based on that
        parent's fitness"""
        parents = []
        parent_repro = self.pop['fitness'].to_numpy()
        parent_repro = np.nan_to_num(parent_repro, copy=True)
        parent_repro /= parent_repro.sum()

        parent_indices = self.pop.index.values
        for n in range(n_parent_pairs):
            p_idx = list(np.random.choice(parent_indices, size=(2,), replace=False, p=parent_repro))
            parents.append((self.pop.iloc[p_idx[0], :], self.pop.iloc[p_idx[1], :]))
        return parents

    def __crossover(self, parent_pairs, n_tries=5):
        """Create a child from each parent pair. Each child gene is uniformly chosen from its parents

        If the child already exists in the current population, new genes are chosen, but maximal n_tries times before a
        ValueError is raised.
        """
        childs = []
        for parents in parent_pairs:
            count = 0
            while True:
                if count > n_tries:
                    raise ValueError
                child_genes = []
                child_sigma = []
                for g, gene in enumerate(self.initial_gene_pool):
                    choice = np.random.uniform()
                    if choice > 0.5:
                        child_genes.append(parents[0][gene])
                        child_sigma.append(parents[0]['sigma'][g])
                    else:
                        child_genes.append(parents[1][gene])
                        child_sigma.append(parents[1]['sigma'][g])
                already_exists = (self.pop.drop(['fitness', 'sigma'], axis=1) == child_genes).all(1).any()
                if not already_exists:
                    break
                count += 1
            childs.append((child_genes, child_sigma))
        return childs

    def plot_genes(self, pop_member):
        for column in pop_member.columns.tolist():
            data = pop_member.loc[:, column].to_list()
            print(f'{column}: ', end="")
            print(*data)

    def eval_fitness(self, target: list, *argv, **kwargs):
        raise NotImplementedError


class GSGeneticAlgorithm(GeneticAlgorithmTemplate):
    from scipy.spatial.distance import cdist

    def __init__(self, gs_config, fitness_measure=cdist):
        super().__init__()

        self.fitness_measure = fitness_measure
        self.gs_config = gs_config

    def eval_fitness(self, target: list, **kwargs):
        param_grid = self.pop.drop(['fitness', 'sigma'], axis=1)

        results = grid_search(circuit_template=self.gs_config['circuit_template'],
                              param_grid=param_grid,
                              param_map=self.gs_config['param_map'],
                              simulation_time=self.gs_config['simulation_time'],
                              dt=self.gs_config['dt'],
                              sampling_step_size=self.gs_config['sampling_step_size'],
                              permute_grid=False,
                              inputs=self.gs_config['inputs'],
                              outputs=self.gs_config['outputs'].copy()
                              )

        for i, candidate_genes in enumerate(param_grid.values):
            candidate_out = results.loc[:, tuple(candidate_genes)].values.T
            target_reshaped = np.array(target)[None, :]
            dist = self.fitness_measure(candidate_out, target_reshaped)
            self.pop.at[i, 'fitness'] = float(1 / dist)


class CGSGeneticAlgorithm(GeneticAlgorithmTemplate):
    def __init__(self, gs_config, cgs_config):
        super().__init__()

        self.gs_config = gs_config
        self.cgs_config = cgs_config

        self.cgs = ClusterGridSearch(cgs_config['nodes'], compute_dir=cgs_config['compute_dir'])

    def eval_fitness(self, target: list, **kwargs):

        param_grid = self.pop.drop(['fitness', 'sigma'], axis=1)

        res_file = self.cgs.run(
            circuit_template=self.gs_config['circuit_template'],
            params=param_grid,
            param_map=self.gs_config['param_map'],
            simulation_time=self.gs_config['simulation_time'],
            dt=self.gs_config['dt'],
            inputs=self.gs_config['inputs'],
            outputs=self.gs_config['outputs'],
            sampling_step_size=self.gs_config['sampling_step_size'],
            permute=False,
            chunk_size=self.cgs_config['chunk_size'],
            worker_env=self.cgs_config['worker_env'],
            worker_file=self.cgs_config['worker_file'],
            config_kwargs={
                'target': target
            })

        results = pd.read_hdf(res_file, key=f'/Results/fitness')

        for i, candidate_genes in enumerate(param_grid.values):
            self.pop.at[i, 'fitness'] = float(results.loc['fitness', tuple(candidate_genes)])







# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
# external imports
import re
from copy import deepcopy
from typing import Union

# pyrates internal imports
from pyrates import PyRatesException
from pyrates.frontend.template.abc import AbstractBaseTemplate
from pyrates.frontend.utility import deep_freeze

# meta infos
from pyrates.ir.operator import OperatorIR

__author__ = " Daniel Rose"
__status__ = "Development"


class OperatorTemplate(AbstractBaseTemplate):
    """Generic template for an operator with a name, equations, variables and possible
    initialization conditions. The template can be used to create variations of a specific
    equations or variables."""

    cache = {}  # tracks all unique instances of applied operator templates
    target_ir = OperatorIR
    key_map = {}
    key_counter = {}

    # key_map = {}  # tracks renaming of of operator keys to have shorter unique keys

    def __init__(self, name: str, path: str, equations: Union[list, str], variables: dict,
                 description: str = "An operator template."):
        """For now: only allow single equations in operator template."""

        super().__init__(name, path, description)

        if isinstance(equations, str):
            self.equations = [equations]
        else:
            self.equations = equations
        self.variables = variables

    def update_template(self, name: str, path: str, equations: Union[str, list, dict] = None,
                        variables: dict = None, description: str = None):
        """Update all entries of the Operator template in their respective ways."""

        if equations:
            # if it is a string, just replace
            if isinstance(equations, str):
                equations = [equations]
            elif isinstance(equations, list):
                pass  # pass equations string to constructor
            # else, update according to predefined rules, assuming dict structure
            elif isinstance(equations, dict):
                equations = [_update_equation(eq, **equations) for eq in self.equations]

            else:
                raise TypeError("Unknown data type for attribute 'equations'.")
        else:
            # copy equations from parent template
            equations = self.equations

        if variables:
            variables = _update_variables(self.variables, variables)
        else:
            variables = self.variables

        rogue_variables = set()
        for var in variables:
            # remove variables that are not present in the equations
            found = False
            for eq in equations:
                if var in eq:
                    found = True

            if not found:
                # save entries in list, since dictionary must not change size during iteration
                rogue_variables.add(var)

        for var in rogue_variables:
            variables.pop(var)

        if not description:
            description = self.__doc__  # or do we want to enforce documenting a template?

        return self.__class__(name=name, path=path, equations=equations, variables=variables,
                              description=description)

    def apply(self, return_key=False, values: dict = None):
        """Returns the non-editable but unique, cashed definition of the operator."""

        # figure out key name of given combination of template and assigned values
        name = self.name
        if values:
            # if values are given, figure out, whether combination is known
            frozen_values = deep_freeze(values)
            try:
                key = self.key_map[(name, frozen_values)]
            except KeyError:
                # not known, need to assign new key and register key in key_map
                if name in self.key_counter:
                    key = f"{name}.{self.key_counter[name] + 1}"
                    self.key_counter[name] += 1
                else:
                    key = f"{name}.1"
                    self.key_counter[name] = 1
                self.key_map[(name, frozen_values)] = key

        else:  # without additional values specified, default name is op_name.0
            key = f"{name}.0"

        try:
            instance, variables = self.cache[key]
            # instance = defrost(instance)
            # variables = defrost(variables)
            # instance, variables = dict(instance), dict(variables)
        except KeyError:
            # get variable definitions and specified default values
            # ToDo: remove variable separation: instead pass variables detached from equations?
            variables, inputs, output = self._separate_variables(self)

            # reduce order of ODE if necessary
            # this step is currently skipped to streamline equation interface
            # instead equations need to be given as either non-differential equations or first-order
            # linear differential equations
            # *equations, variables = cls._reduce_ode_order(template.equations, variables)
            equations = self.equations

            # operator instance is invoked as a dictionary of equations and variable definition
            # this may be subject to change

            if values:
                for vname, update in values.items():
                    # if isinstance(update, dict):
                    #     if "value" in update:
                    #         variables[vname["value"]] = update["value"]
                    #     if "vtype" in
                    # only values are updated, assuming all other specs remain the same
                    variables[vname]["value"] = values[vname]
                    # should fail, if variable is unknown

            instance = self.target_ir(equations=equations, inputs=inputs, output=output)
            self.cache[key] = (instance, variables)

        if return_key:
            return instance.copy(), deepcopy(variables), key
        else:
            return instance.copy(), deepcopy(variables)

    @classmethod
    def _separate_variables(cls, template):
        """
        Return variable definitions and the respective values.

        Returns
        -------
        variables
        inputs
        output
        """
        # this part can be improved a lot with a proper expression parser

        variables = {}
        inputs = {}
        output = None
        for variable, properties in template.variables.items():
            var_dict = deepcopy(properties)
            # default shape is scalar
            if "shape" not in var_dict:
                var_dict["shape"] = (1,)

            # identify variable type and data type
            # note: this assume that a "default" must be given for every variable
            try:
                var_dict.update(cls._parse_vprops(var_dict.pop("default")))
            except KeyError:
                raise PyRatesException("Variables need to have a 'default' (variable type, data type and/or value) "
                                       "specified.")

            # separate in/out specification from variable type specification
            if var_dict["vtype"] == "input":
                inputs[variable] = dict(sources=[], reduce_dim=True)  # default to True for now
                var_dict["vtype"] = "state_var"
            elif var_dict["vtype"] == "output":
                if output is None:
                    output = variable  # for now assume maximum one output is present
                else:
                    raise PyRatesException("More than one output specification found in operator. "
                                           "Only one output per operator is supported.")
                var_dict["vtype"] = "state_var"

            variables[variable] = var_dict

        return variables, inputs, output

    @staticmethod
    def _parse_vprops(expr: Union[str, int, float]):
        """Naive version of a parser for the default key of variables in a template. Returns data type,
        variable type and default value of the variable."""

        value = 0.  # Setting initial conditions for undefined variables to 0. Is that reasonable?
        if isinstance(expr, int):
            vtype = "constant"
            value = expr
            # dtype = "int32"
            dtype = "float32"  # default to float for maximum compatibility
        elif isinstance(expr, float):
            vtype = "constant"
            value = expr
            dtype = "float32"
            # restriction to 32bit float for consistency. May not be reasonable at all times.
        else:
            # set vtype
            if expr.startswith("input"):
                vtype = "input"
            elif expr.startswith("output"):
                vtype = "output"
            elif expr.startswith("variable"):
                vtype = "state_var"
            elif expr.startswith("constant"):
                vtype = "constant"
            elif expr.startswith("placeholder"):
                vtype = "placeholder"
            else:
                try:
                    # if "." in expr:
                    value = float(expr)  # default to float
                    # else:
                    #     value = int(expr)
                    vtype = "constant"
                except ValueError:
                    raise ValueError(f"Unable to interpret variable type in default definition {expr}.")

            # set dtype and value
            if expr.endswith("(float)"):
                dtype = "float32"  # why float32 and not float64?
            elif expr.endswith("(int)"):
                # dtype = "int32"
                dtype = "float32"  # default to float for maximum compatibility
            elif "." in expr:
                dtype = "float32"
                value = float(re.search("[+-]?([0-9]+([.][0-9]*)?|[.][0-9]+)", expr).group())
                # see https://stackoverflow.com/questions/12643009/regular-expression-for-floating-point-numbers
            elif re.search("[0-9]+", expr):
                # dtype = "int32"
                dtype = "float32"  # default to float for maximum compatibility
                value = float(re.search("[0-9]+", expr).group())
            else:
                dtype = "float32"  # base assumption

        return dict(vtype=vtype, dtype=dtype, value=value)


def _update_variables(variables: dict, updates: dict):
    updated = deepcopy(variables)

    for var, var_dict in updates.items():
        if var in updated:
            # update dictionary defining single variable
            updated[var].update(var_dict)
        else:
            # copy new variable into variables dictionary
            updated.update({var: var_dict})

    return updated


def _update_equation(equation: str,  # original equation
                     replace: dict = False,  # replace parts of the string
                     remove: Union[list, tuple] = False,  # remove parts of the string
                     append: str = False,  # append to the end of the string
                     prepend: str = False,  # add to beginning of string
                     ):
    # replace existing terms by new ones
    if replace:
        for old, new in replace.items():
            equation = equation.replace(old, new)
            # this might fail, if multiple replacements refer or contain the same variables
            # is it possible to call str.replace with tuples?

    # remove terms
    if remove:
        if isinstance(remove, str):
            equation = equation.replace(remove, "")
        else:
            for old in remove:
                equation = equation.replace(old, "")

    # append terms at the end of the equation string
    if append:
        # only allowing single append per update
        equation = f"{equation} {append}"

        # prepend terms at the beginning of the equation string
    if prepend:
        # only allowing single prepend per update
        equation = f"{prepend} {equation}"

    return equation




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
from copy import deepcopy

import networkx as nx
from networkx import MultiDiGraph, DiGraph, find_cycle, NetworkXNoCycle

from pyrates import PyRatesException
from pyrates.ir.edge import EdgeIR
from pyrates.ir.circuit import CircuitIR
from pyrates.frontend.dict import to_node
from pyrates.frontend._registry import register_interface


__author__ = "Daniel Rose"
__status__ = "Development"


@register_interface
def to_circuit(graph: nx.MultiDiGraph, label="circuit",
               node_creator=to_node):
    """Create a CircuitIR instance out of a networkx.MultiDiGraph"""

    circuit = CircuitIR(label)

    for name, data in graph.nodes(data=True):

        circuit.add_node(name, node=node_creator(data))

    required_keys = ["source_var", "target_var", "weight", "delay"]
    for source, target, data in graph.edges(data=True):

        if all([key in data for key in required_keys]):
            if "edge_ir" not in data:
                data["edge_ir"] = EdgeIR()
            source_var = data.pop("source_var")
            target_var = data.pop("target_var")
            circuit.add_edge(f"{source}/{source_var}", f"{target}/{target_var}", **data)
        else:
            raise KeyError(f"Missing a key out of {required_keys} in an edge with source `{source}` and target"
                           f"`{target}`")

    return circuit


def from_circuit(circuit, revert_node_names=False):
    """Old implementation that transforms all information in a circuit to a networkx.MultiDiGraph with a few additional
    transformations that the old backend needed."""
    return Circuit2NetDef.network_def(circuit, revert_node_names)


class Circuit2NetDef:
    # label_counter = {}  # type: Dict[str, int]

    @classmethod
    def network_def(cls, circuit: CircuitIR, revert_node_names=False):
        """A bit of a workaround to connect interfaces of frontend and backend.
        TODO: Remove BackendIRFormatter and adapt corresponding tests"""
        # import re

        network_def = MultiDiGraph()

        edge_list = []
        node_dict = {}

        # reorganize node to conform with backend API
        #############################################
        for node_key, data in circuit.graph.nodes(data=True):
            node = data["node"]
            # reformat all node internals into operators + operator_args
            if revert_node_names:
                names = node_key.split("/")
                node_key = ".".join(reversed(names))
            node_dict[node_key] = {}  # type: Dict[str, Union[list, dict]]
            node_dict[node_key] = dict(cls._nd_reformat_operators(node.op_graph))
            op_order = cls._nd_get_operator_order(node.op_graph)  # type: list
            # noinspection PyTypeChecker
            node_dict[node_key]["operator_order"] = op_order

        # reorganize edge to conform with backend API
        #############################################
        for source, target, data in circuit.graph.edges(data=True):
            # move edge operators to node
            if revert_node_names:
                source = ".".join(reversed(source.split("/")))
                target = ".".join(reversed(target.split("/")))
            node_dict[target], edge = cls._move_edge_ops_to_node(target, node_dict[target], data)

            edge_list.append((source, target, dict(**edge)))

        # network_def.add_nodes_from(node_dict)
        for key, node in node_dict.items():
            network_def.add_node(key, **node)
        network_def.add_edges_from(edge_list)

        return network_def  # return MultiDiGraph as needed by ComputeGraph class

    @staticmethod
    def _nd_reformat_operators(op_graph: DiGraph):
        operator_args = dict()
        operators = dict()

        for op_key, op_dict in op_graph.nodes(data=True):
            op_cp = deepcopy(op_dict)  # duplicate operator info
            var_dict = op_cp.pop("variables")

            for var_key, var_props in var_dict.items():
                var_prop_cp = deepcopy(var_props)  # duplicate variable properties

                # if var_key in op_dict["variables"]:  # workaround to get values back into operator_args
                #     var_prop_cp["value"] = deepcopy(op_dict["values"][var_key])

                var_prop_cp["shape"] = ()  # default to scalars for now
                var_prop_cp.pop("unit", None)
                var_prop_cp.pop("description", None)
                var_prop_cp.pop("name", None)
                # var_prop_cp["name"] = f"{new_op_key}/{var_key}"  # has been thrown out
                operator_args[f"{op_key}/{var_key}"] = deepcopy(var_prop_cp)

            op_cp["equations"] = op_cp["operator"].equations
            op_cp["inputs"] = op_cp["operator"].inputs
            op_cp["output"] = op_cp["operator"].output
            # op_cp.pop("values", None)
            op_cp.pop("operator", None)
            operators[op_key] = op_cp

        reformatted = dict(operator_args=operator_args,
                           operators=operators,
                           inputs={})
        return reformatted

    @staticmethod
    def _nd_get_operator_order(op_graph: DiGraph) -> list:
        """

        Parameters
        ----------
        op_graph

        Returns
        -------
        op_order
        """
        # check, if cycles are present in operator graph (which would be problematic
        try:
            find_cycle(op_graph)
        except NetworkXNoCycle:
            pass
        else:
            raise PyRatesException("Found cyclic operator graph. Cycles are not allowed for operators within one node.")

        op_order = []
        graph = op_graph.copy()  # type: DiGraph
        while graph.nodes:
            # noinspection PyTypeChecker
            primary_nodes = [node for node, in_degree in graph.in_degree if in_degree == 0]
            op_order.extend(primary_nodes)
            graph.remove_nodes_from(primary_nodes)

        return op_order

    @classmethod
    def _move_edge_ops_to_node(cls, target, node_dict: dict, edge_dict: dict) -> (dict, dict):
        """

        Parameters
        ----------
        target
            Key identifying target node in backend graph
        node_dict
            Dictionary of target node (to move operators into)
        edge_dict
            Dictionary with edge properties (to move operators from)
        Returns
        -------
        node_dict
            Updated dictionary of target node
        edge_dict
             Dictionary of reformatted edge
        """
        # grab all edge variables
        edge = edge_dict["edge_ir"]  # type: EdgeIR
        source_var = edge_dict["source_var"]
        target_var = edge_dict["target_var"]
        weight = edge_dict["weight"]
        delay = edge_dict["delay"]
        input_var = edge.input
        output_var = edge.output

        if len(edge.op_graph) > 0:
            # reformat all edge internals into operators + operator_args
            op_data = cls._nd_reformat_operators(edge.op_graph)  # type: dict
            op_order = cls._nd_get_operator_order(edge.op_graph)  # type: List[str]
            operators = op_data["operators"]
            operator_args = op_data["operator_args"]

            # operator keys refer to a unique combination of template names and changed values

            # add operators to target node in reverse order, so they can be safely prepended
            added_ops = False
            for op_name in reversed(op_order):
                # check if operator name is already known in target node
                if op_name in node_dict["operators"]:
                    pass
                else:
                    added_ops = True
                    # this should all go smoothly, because operator should not be known yet
                    # add operator dict to target node operators
                    node_dict["operators"][op_name] = operators[op_name]
                    # prepend operator to op_order
                    node_dict["operator_order"].insert(0, op_name)
                    # ToDo: consider using collections.deque instead
                    # add operator args to target node
                    node_dict["operator_args"].update(operator_args)

            out_op = op_order[-1]
            out_var = operators[out_op]['output']
            if added_ops:
                # append operator output to target operator sources
                # assume that only last operator in edge operator_order gives the output
                # for op_name in node_dict["operators"]:
                #     if out_var in node_dict["operators"][op_name]["inputs"]:
                #         if out_var_long not in node_dict["operators"][op_name]["inputs"][out_var]:
                #             # add reference to source operator that was previously in an edge
                #             node_dict["operators"][op_name]["inputs"][out_var].append(output_var)

                # shortcut, since target_var and output_var are known:
                target_op, target_vname = target_var.split("/")
                if output_var not in node_dict["operators"][target_op]["inputs"][target_vname]["sources"]:
                    node_dict["operators"][target_op]["inputs"][target_vname]["sources"].append(out_op)

            # simplify edges and save into edge_list
            # op_graph = edge.op_graph
            # in_ops = [op for op, in_degree in op_graph.in_degree if in_degree == 0]
            # if len(in_ops) == 1:
            #     # simple case: only one input operator? then it's the first in the operator order.
            #     target_op = op_order[0]
            #     target_inputs = operators[target_op]["inputs"]
            #     if len(target_var) != 1:
            #         raise PyRatesException("Either too many or too few input variables detected. "
            #                                "Needs to be exactly one.")
            #     target_var = list(target_inputs.keys())[0]
            #     target_var = f"{target_op}/{target_var}"
            # else:
            #     raise NotImplementedError("Transforming an edge with multiple input operators is not yet handled.")

            # shortcut to new target war:
            target_var = input_var
        edge_dict = {"source_var": source_var,
                     "target_var": target_var,
                     "weight": weight,
                     "delay": delay}
        # set target_var to singular input of last operator added
        return node_dict, edge_dict



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""Functions for performing parameter grid simulations with pyrates model_templates.
"""

# external imports
import pandas as pd
import numpy as np
from typing import Optional, Union

# system imports
import os
import sys
import json
import time as t
import glob
import getpass
from shutil import copy2
from pathlib import Path
from datetime import datetime
from threading import Thread, currentThread, RLock

# pyrates internal imports
from pyrates.backend import ComputeGraph
from pyrates.frontend import CircuitTemplate
from pyrates.ir.circuit import CircuitIR

# meta infos
__author__ = "Christoph Salomon, Richard Gast"
__status__ = "development"


def grid_search(circuit_template: Union[CircuitTemplate, str], param_grid: Union[dict, pd.DataFrame], param_map: dict, dt: float, simulation_time: float,
                inputs: dict, outputs: dict, sampling_step_size: Optional[float] = None,
                permute_grid: bool = False, init_kwargs: dict = None, **kwargs) -> pd.DataFrame:
    """Function that runs multiple parametrizations of the same circuit in parallel and returns a combined output.

    Parameters
    ----------
    circuit_template
        Path to the circuit template.
    param_grid
        Key-value pairs for each circuit parameter that should be altered over different circuit parametrizations.
    param_map
        Key-value pairs that map the keys of param_grid to concrete circuit variables.
    dt
        Simulation step-size in s.
    simulation_time
        Simulation time in s.
    inputs
        Inputs as provided to the `run` method of `:class:ComputeGraph`.
    outputs
        Outputs as provided to the `run` method of `:class:ComputeGraph`.
    sampling_step_size
        Sampling step-size as provided to the `run` method of `:class:ComputeGraph`.
    permute_grid
        If true, all combinations of the provided param_grid values will be realized. If false, the param_grid values
        will be traversed pairwise.
    kwargs
        Additional keyword arguments passed to the `:class:ComputeGraph` initialization.



    Returns
    -------
    pd.DataFrame
        Simulation results stored in a multi-index data frame where each index lvl refers to one of the parameters of
        param_grid.

    """

    # load template if necessary
    if type(circuit_template) is str:
        circuit_template = CircuitTemplate.from_yaml(circuit_template)

    # linearize parameter grid if necessary
    if type(param_grid) is dict:
        param_grid = linearize_grid(param_grid, permute_grid)

    # create multi-index for later results storage
    param_keys = list(param_grid.keys())
    multi_idx = [param_grid[key].values for key in param_keys]
    n_iters = len(multi_idx[0])
    outs = []
    out_names = list(outputs.keys())
    for out_name in out_names:
        outs += [out_name] * n_iters
    multi_idx = [list(idx) * len(out_names) for idx in multi_idx]
    multi_idx = multi_idx + [outs]
    index = pd.MultiIndex.from_arrays(multi_idx, names=param_keys + ["out_var"])
    index = pd.MultiIndex.from_tuples(list(set(index)), names=param_keys + ["out_var"])

    # assign parameter updates to each circuit and combine them to unconnected network
    circuit = CircuitIR()
    circuit_names = []
    param_keys = index.names
    results_indices = []
    i = 0
    for idx in index.values:
        if idx[:-1] not in results_indices:
            results_indices.append(idx[:-1])
            new_params = {}
            for key, val in zip(param_keys, idx):
                if key in param_grid:
                    new_params[key] = val
            circuit_tmp = circuit_template.apply()
            circuit_names.append(f'{circuit_tmp.label}_{i}')
            circuit_tmp = adapt_circuit(circuit_tmp, new_params, param_map)
            circuit.add_circuit(circuit_names[-1], circuit_tmp)
            i += 1

    # create backend graph
    if not init_kwargs:
        init_kwargs = {}
    vectorization = init_kwargs.pop('vectorization', 'nodes')
    net = ComputeGraph(circuit, dt=dt, vectorization=vectorization, **init_kwargs)

    # adjust input of simulation to combined network
    for inp_key, inp in inputs.items():
        inputs[inp_key] = np.tile(inp, (1, len(circuit_names)))

    # adjust output of simulation to combined network
    nodes = list(circuit_template.apply().nodes)
    out_nodes = []
    for out_key in out_names:
        out = outputs.pop(out_key)
        out = out.split('/')
        out_nodes.append(out)
        node = "/".join(out[:-2])
        if node in nodes:
            op, var = out[-2], out[-1]
            node = node.split('.')[0]
            outputs[out_key] = "/".join((node, op, var))

    # simulate the circuits behavior
    results = net.run(simulation_time=simulation_time,
                      inputs=inputs,
                      outputs=outputs,
                      sampling_step_size=sampling_step_size,
                      **kwargs)    # type: pd.DataFrame
    if 'profile' in kwargs:
        results, duration, memory = results

    # transform results into long-form dataframe with changed parameters as columns
    results_final = pd.DataFrame(columns=index, data=np.zeros_like(results.values), index=results.index)
    for out_name, out_info in zip(out_names, out_nodes):
        node, op, var = out_info
        for node_name, params in zip(circuit_names, results_indices):
            key = params + (out_name,)
            idx = list(net.get_var(f"{node_name}/{node}", op, var, retrieve=False).values())
            results_final[key] = results[out_name].iloc[:, idx]

    if 'profile' in kwargs:
        return results_final, duration, memory
    return results_final


class ClusterCompute:
    def __init__(self, nodes: list, compute_dir=None):
        """Create new ClusterCompute instance object with unique compute ID

        Creates a compute directory for the ClusterCompute instance, either in the specified path or as a default
        folder in the current working directory of the executing script.
        Creates a logfile in ComputeDirectory/Logs and tees all future stdout and stderr of the executing script
        to this file
        Connects to all nodes via SSH and saves the corresponding paramiko client in self.clients

        Parameters
        ----------
        nodes:
            List of names or IP addresses of working stations/servers in the local network
        compute_dir:
            Full path to a directory that will be used as compute directory.
            If none is given, a default compute directory is created in the current working directory

        Returns
        -------
        ClusterCompute instance object

        """

        self.clients = []
        self.lock = RLock()

        #############################################################
        # Create main compute directory, subdirectories and logfile #
        #############################################################

        # Unique compute ID based on date and time
        self.compute_id = datetime.now().strftime("%d%m%y-%H%M%S")

        # Main compute directory
        if compute_dir:
            self.compute_dir = compute_dir
            os.makedirs(self.compute_dir, exist_ok=True)
        else:
            # Create default compute directory in the current working directory
            self.compute_dir = f'{os.getcwd()}/ClusterCompute_{self.compute_id}'
            os.makedirs(self.compute_dir, exist_ok=True)

        # Logfile directory
        self.log_dir = f'{self.compute_dir}/Logs/{self.compute_id}'
        os.makedirs(self.log_dir, exist_ok=True)

        # Global logfile to copy stdout and stderr to
        self.global_logfile = f'{self.log_dir}/Global_logfile.log'
        os.makedirs(os.path.dirname(self.global_logfile), exist_ok=True)

        ####################################
        # Tee stdout and stderr to logfile #
        ####################################
        sys.stdout = StreamTee(sys.stdout, self.global_logfile)
        sys.stderr = StreamTee(sys.stderr, self.global_logfile)

        ###########################
        # Write header to logfile #
        ###########################
        print("***NEW CLUSTER INSTANCE CREATED***")
        print(f'Compute ID: {self.compute_id}')
        print(f'Compute directory: {self.compute_dir}')
        print(f'Global logfile: {self.global_logfile}')
        print("")

        ##################
        # Create cluster #
        ##################
        print("***CONNECTING TO NODES...***")
        t0 = t.time()
        self.cluster_connect(nodes)
        print(f'Nodes connected. Elapsed time: {t.time()-t0:.3f} seconds')

    def __del__(self):
        # Make sure to close all clients when cluster_examples instance is destroyed
        for client in self.clients:
            try:
                client["paramiko_client"].close()
            except TypeError:
                pass

    def cluster_connect(self, nodes: list):
        """Create SSH connections to all nodes in the list, respectively

        Connect to all nodes in the given list via SSH, using ssh_connect(). Adds a dictionary for each node to the
        class internal 'clients' list
        Each dictionary contains:
            - ["paramiko_client"]: A paramiko client that can be used to execute commands on the node
            - ["node_name"]: The computer name of the connected node
            - ["hardware"]: A dictionary with certain hardware information of the node
            - ["logfile"]: Path to a logfile inside the instance's working directory,
                           where all stdout and stderr of the node will be redirected to

        Parameters
        ----------
        nodes
            List with names or IP addresses of working stations/servers in the local network

        Returns
        -------

        """
        username = getpass.getuser()
        # password = getpass.getpass(prompt='Enter password: ', stream=sys.stderr)
        for node in nodes:
            client = self.ssh_connect(node, username=username)
            if client is not None:
                # Create local logfiles
                local_logfile = f'{self.log_dir}/Local_logfile_{node}.log'
                os.makedirs(os.path.dirname(local_logfile), exist_ok=True)

                hw = self.get_hardware_spec(client)

                self.clients.append({
                    "paramiko_client": client,
                    "node_name": node,
                    "hardware": hw,
                    "logfile": local_logfile})

    def spawn_thread(self, client, thread_kwargs):
        """

        Parameters
        ----------
        client
        thread_kwargs

        Returns
        -------

        """
        t_ = Thread(
            name=client["node_name"],
            target=self.thread_master,
            args=(client, thread_kwargs)
        )
        t_.start()
        return t_

    def thread_master(self, client, thread_kwargs_: dict):
        """Function that is executed by every thread. Every instance of thread_master is bound to a different client

        The thread_master() function can be arbitrarily changed to fit the needs of the user.
        Commandline commands on the remote node can be executed using client["paramiko_client"].exec_command("")
        self.lock can be used to ensure that a code snippet is executed without threads being switched in between
        e.g.
            with self.lock:
                some code that is executed without switching to another thread
        Since thread_master() is called by spawn_threads(), which again is called by run(), **kwargs of
        spawn_threads() are parsed as a single dict to thread_master(). **kwargs of run() are NOT automatically parsed
        to spawn_threads() or thread_master().

        Params
        ------
        client
            dict containing a paramiko client, name of the connected node, dict with hardware specifications and
            a path to a logfile.
        kwargs_
            dict containing **kwargs of spawn_threads() as key/value pairs

        Returns
        -------

        """
        pass

    @staticmethod
    def get_hardware_spec(pm_client):
        """Print hardware specifications of the workstation tied to the paramiko client

        Prints CPU model, number of CPU cores, min CPU freq, max CPU freq and working memory of a paramiko client
        connected to a remote worker node

        Parameters
        ----------
        pm_client
            Paramiko client

        Returns
        -------
        dict


        """

        stdin, stdout, stderr = pm_client.exec_command("lscpu | grep 'Model name'")
        cpu = stdout.readline().split()
        cpu = ' '.join(map(str, cpu[2:]))
        print(f'CPU: {cpu}')

        stdin, stdout, stderr = pm_client.exec_command("lscpu | grep 'CPU(s)'")
        num_cpu_cores = int(stdout.readline().split(":")[1])
        print(f'Cores: {num_cpu_cores}')

        stdin, stdout, stderr = pm_client.exec_command("lscpu | grep 'min MHz'")
        cpu_min = float(stdout.readline().split(":")[1])
        print(f'CPU min: {cpu_min} MHz')

        stdin, stdout, stderr = pm_client.exec_command("lscpu | grep 'max MHz'")
        cpu_max = float(stdout.readline().split(":")[1])
        print(f'CPU max: {cpu_max} MHz')

        stdin, stdout, stderr = pm_client.exec_command("free -m | grep 'Mem'")
        total_mem = int(stdout.readline().split()[1])
        print(f'Total memory: {total_mem} MByte')
        print("")

        hw = {
            'cpu': cpu,
            'num_cpu_cores': num_cpu_cores,
            'cpu_min': cpu_min,
            'cpu_max': cpu_max,
            'total_mem': total_mem
        }
        return hw

    @staticmethod
    def ssh_connect(node, username, password=None, print_status=True):
        """Connect to a host via SSH and return a respective paramiko.SSHClient

        Parameters
        ----------
        node
            Name or IP-address of the host to connect to
        username
        password
            ssh_connect uses kerberos authentication so no password is required. If kerberos is not available on your
            network, Paramiko needs a password to create the SSH connection
        Returns
        -------
        paramiko.SSHClient()
            Is None if connection fails. For detailed information of the thrown exception see Paramiko documentation

        """
        import paramiko
        client = paramiko.SSHClient()
        try:
            # Using kerberos authentication
            client.connect(node, username=username, gss_auth=True, gss_kex=True)
            if print_status:
                print(f'\'{node}\': Connection established!')
            return client
        except paramiko.ssh_exception.NoValidConnectionsError as err:
            if print_status:
                print(f'\'{node}\': ', err)
            return None
        except paramiko.ssh_exception.AuthenticationException as err:
            if print_status:
                print(f'\'{node}\': ', err)
            return None
        except paramiko.ssh_exception.SSHException as err:
            if print_status:
                print(f'\'{node}\': ', err)
            return None
        except IOError as err:
            if print_status:
                print(f'\'{node}\': ', err)
            return None


class ClusterGridSearch(ClusterCompute):
    def __init__(self, nodes, compute_dir=None):
        super().__init__(nodes, compute_dir)

        self.chunk_idx = 0
        self.res_file_collection = {}

        # Adding additional subfolders to the compute directory
        # Grid directory
        self.grid_dir = f'{self.compute_dir}/Grids'
        os.makedirs(self.grid_dir, exist_ok=True)

        # Config directory
        self.config_dir = f'{self.compute_dir}/Config'
        os.makedirs(self.config_dir, exist_ok=True)

        # Result directory
        self.res_dir = f'{self.compute_dir}/Results'
        os.makedirs(self.res_dir, exist_ok=True)

    def run(self, circuit_template: str, params: Union[dict, pd.DataFrame], param_map: dict, dt: float, simulation_time: float,
            inputs: dict, outputs: dict, sampling_step_size: float, chunk_size: (int, list),
            worker_env: str, worker_file: str, result_kwargs: dict = {}, config_kwargs: dict = {},
            add_template_info: bool = False, permute: bool = False, **kwargs) -> str:
        """Run multiple instances of grid_search simultaneously on different workstations in the compute cluster

        ClusterGridSearch requires all necessary data (worker script, worker environment, compute directory)
        to be stored on a server. After a connection is established, all data on the remote workers has to be accessible
        in the same way as it would be on the master worker (same root directory, same filepaths, etc.)
        Paths inside the run function have to be adjusted respectively.

        FTP implementation is planned in future versions for direct file transfer between the master and the workers.

        Parameters
        ----------
        circuit_template
            Path to the circuit template.
        params
            Dictionary containing lists of parameters to create the parameter grid from.
        param_map
            Key-value pairs that map the keys of param_grid to circuit variables.
        dt
            Simulation step-size in s.
        simulation_time
            Simulation time in s.
        inputs
            inputs as provided to the `run` method of `:class:ComputeGraph`.
        outputs
            Outputs as provided to the `run` method of `:class:ComputeGraph`.
        sampling_step_size
            Sampling step-size as provided to the `run` method of `:class:ComputeGraph`.
        permute
            If true, all combinations of the parameter values in params will be created.
        chunk_size
            Number of parameter combinations computed simultaneously on one worker
        worker_env
            Python executable inside an environment in which the remote worker script is called.
        worker_file
            Python script that will be executed by each remote worker.
        result_kwargs
            Key-value pairs that will be added to the result file's 'AdditionalData' dataset
        config_kwargs
            Key-value pairs that will be added to the config file.
        add_template_info
            If true, all operator templates and its variables are copied from the yaml file to a dedicated folder in the
            result file

        Returns
        -------
        str
            .hdf5 file containing the computation results as DataFrame in dataset '/Results/...'
        """

        import h5py
        t_total = t.time()

        print("")

        # Prepare parameter grid
        ########################
        print("***PREPARING PARAMETER GRID***")
        t0 = t.time()

        # Create DataFrame from param dictionary
        if isinstance(params, dict):
            param_grid = linearize_grid(params, permute=permute)
        else:
            param_grid = params

        # Create default parameter grid csv-file
        grid_idx = 0
        while os.path.exists(f'{self.grid_dir}/DefaultGrid_{grid_idx}.csv'):
            grid_idx += 1
        grid_file = f'{self.grid_dir}/DefaultGrid_{grid_idx}.csv'
        grid_name = Path(grid_file).stem
        param_grid.to_csv(grid_file, index=True)

        # Add desired chunk size to each node
        #####################################
        for i, client in enumerate(self.clients):
            if isinstance(chunk_size, list):
                client['chunk_size'] = chunk_size[i]
            else:
                client['chunk_size'] = chunk_size

        # Assign chunk indices to parameter grid
        working_grid = param_grid.copy()
        working_grid["status"] = 'unsolved'
        working_grid["chunk_idx"] = -1
        working_grid["err_count"] = 0

        print(f'Done! Elapsed time: {t.time() - t0:.3f} seconds')

        print("")

        # Check key consistency
        #######################
        print("***CHECKING PARAMETER GRID AND MAP FOR CONSISTENCY***")
        t0 = t.time()
        if not self.check_key_consistency(params, param_map):
            print("Terminating execution!")
            return ""
        print(f'Done! Elapsed time: {t.time() - t0:.3f} seconds')

        print("")

        # Create config file
        #####################
        print("***CREATING CONFIG FILE***")
        t0 = t.time()
        config_idx = 0
        while os.path.exists(f'{self.config_dir}/DefaultConfig_{config_idx}.json'):
            config_idx += 1
        config_file = f'{self.config_dir}/DefaultConfig_{config_idx}.json'
        self.create_cgs_config(fp=config_file,
                               circuit_template=circuit_template,
                               param_map=param_map,
                               dt=dt,
                               simulation_time=simulation_time,
                               inputs=inputs,
                               outputs=outputs,
                               sampling_step_size=sampling_step_size,
                               add_kwargs=config_kwargs)
        print(f'Done! Elapsed time: {t.time() - t0:.3f} seconds')

        print("")

        # Create global result file
        ############################
        print("***CREATING GLOBAL RESULT FILE***")
        t0 = t.time()
        # Create result directory and result file for current parameter grid
        grid_res_dir = f'{self.res_dir}/{grid_name}'
        os.makedirs(grid_res_dir, exist_ok=True)
        global_res_file = f'{grid_res_dir}/CGS_result_{grid_name}.h5'

        # Write grid and config information to global result file
        with h5py.File(global_res_file, 'a') as file:
            for key, value in params.items():
                file.create_dataset(f'/ParameterGrid/Keys/{key}', data=value)
            for key, value in result_kwargs.items():
                file.create_dataset(f'/AdditionalData/{key}', data=value)
            file.create_dataset(f'/Config/config_file', data=config_file)
            file.create_dataset(f'/Config/circuit_template', data=circuit_template)
            file.create_dataset(f'/Config/simulation_time', data=simulation_time)
            file.create_dataset(f'/Config/dt', data=dt)
            file.create_dataset(f'/Config/sampling_step_size', data=sampling_step_size)
            if add_template_info:
                template_file = f'{circuit_template.rsplit("/", 1)[0]}.yaml'
                self.add_template_information(template_file, file)

        param_grid.to_hdf(global_res_file, key='/ParameterGrid/Grid_df')
        print(f'Done. Elapsed time: {t.time() - t0:.3f} seconds')

        print("")

        # Create keyword dictionary for threads
        #######################################
        thread_kwargs = {
            "worker_env": worker_env,
            "worker_file": worker_file,
            "working_grid": working_grid,
            "grid_name": grid_name,
            "grid_res_dir": grid_res_dir,
            "config_file": config_file,
            "global_res_file": global_res_file
        }

        # TODO: Copy worker environment, worker file and configuration file to the worker, if necessary

        # Start cluster computation
        ###########################
        print("***STARTING CLUSTER COMPUTATION***")
        # Spawn threads to control each node connection and start computation
        threads = [self.spawn_thread(client, thread_kwargs) for client in self.clients]
        # Wait for all threads to finish
        for t_ in threads:
            t_.join()

        print("")
        print(f'Cluster computation finished. Elapsed time: {t.time() - t_total:.3f} seconds')

        # Check parameter combinations that failed multiple times to compute
        remaining_params = working_grid.loc[working_grid["status"] == "failed"].index
        if len(remaining_params) > 0:
            print("WARNING: The following parameter indices could not be computed: ")
            print(remaining_params)
        print("")

        # Write local results to global result file
        ###########################################
        print(f'***WRITING RESULTS TO GLOBAL RESULT FILE***')
        t0 = t.time()

        # Get sorted list of temporary result files to iterate through
        temp_res_files = glob.glob(f'{grid_res_dir}/*_temp*')
        temp_res_files.sort()

        # Read number of different circuit outputs and prepare lists to concatenate results
        res_dict = {}
        with pd.HDFStore(temp_res_files[0], "r") as store:
            for key in store.keys():
                res_dict[key] = []

        # Concatenate results from each temporary result file
        for file in temp_res_files:
            with pd.HDFStore(file, "r") as store:
                for idx, key in enumerate(store.keys()):
                    res_dict[key].append(store[key])

        # Create DataFrame for each output variable and write to global result file
        with pd.HDFStore(global_res_file, "a") as store:
            for key, value in res_dict.items():
                if len(value) > 0:
                    df = pd.concat(value, axis=1)
                    store.put(key=f'/Results/{key}', value=df)

        # Delete temporary local result files
        for file in temp_res_files:
            os.remove(file)

        print(f'Elapsed time: {t.time()-t0:.3f} seconds')
        print(f'Find results in: {grid_res_dir}')
        print("")

        working_grid.to_csv(f'{os.path.dirname(grid_file)}/{grid_name}_{self.compute_id}_ResultStatus.csv', index=True)

        # self.__del__()
        return global_res_file

    def thread_master(self, client: dict, thread_kwargs: dict):
        """Function that is executed by each thread to schedule computations on the respective worker

        Parameters
        ----------
        client
            Dictionary containing all information about the remote worker that is tied to the current thread
        thread_kwargs
            Dictionary containing all kwargs that are passed to the thread function

        Returns
        -------

        """

        import paramiko

        thread_name = currentThread().getName()
        connection_lost = False
        connection_lost_counter = 0

        # Get client information
        pm_client = client["paramiko_client"]
        logfile = client["logfile"]
        chunk_size = client["chunk_size"]

        # Get keyword arguments
        config_file = thread_kwargs["config_file"]
        working_grid = thread_kwargs["working_grid"]
        grid_name = thread_kwargs["grid_name"]
        grid_res_dir = thread_kwargs["grid_res_dir"]
        worker_env = thread_kwargs["worker_env"]
        worker_file = thread_kwargs["worker_file"]

        # Prepare worker command
        command = f'{worker_env} {worker_file}'

        # Create folder to save local subgrids to
        subgrid_dir = f'{self.grid_dir}/Subgrids/{grid_name}/{thread_name}'
        os.makedirs(subgrid_dir, exist_ok=True)
        subgrid_idx = 0

        # Start scheduler
        #################
        while True:

            # Temporarily disable thread switching
            with self.lock:

                # Fetch grid chunks
                ####################

                # Get all parameters that haven't been successfully computed yet
                remaining_params = working_grid.loc[working_grid["status"] == "unsolved"]
                if remaining_params.empty:
                    print(f'[T]\'{thread_name}\': No more parameter combinations available!')
                    break
                else:
                    # Find chunk index of first value in remaining params and fetch all parameter combinations with the
                    # same chunk index
                    param_idx, chunk_idx = self._fetch_index(remaining_params, chunk_size, working_grid)
                    working_grid.loc[param_idx, "status"] = "pending"

                    print(f'[T]\'{thread_name}\': Fetching {len(param_idx)} indices: ', end="")
                    print(f'[{param_idx[0]}] - [{param_idx[-1]}]')

                    # Create parameter sub-grid
                    ###########################
                    subgrid_fp = f'{subgrid_dir}/{thread_name}_Subgrid_{subgrid_idx}.h5'
                    subgrid_df = working_grid.iloc[param_idx]
                    subgrid_df.to_hdf(subgrid_fp, key="subgrid")
                    subgrid_idx += 1

                    # TODO: Copy subgrid to worker if necessary

                    # Temporary result file for current subgrid
                    ###########################################
                    local_res_file = f'{grid_res_dir}/CGS_result_{grid_name}_' \
                        f'chunk_{chunk_idx:02}_idx_{param_idx[0]}-{param_idx[-1]}_temp.h5'
                    self.res_file_collection[chunk_idx] = local_res_file

                    # Execute worker script on the remote host
                    ##########################################
                    t0 = t.time()

                    try:
                        print(f'[T]\'{thread_name}\': Starting remote computation...')
                        stdin, stdout, stderr = pm_client.exec_command(command +
                                                                       f' --config_file={config_file}'
                                                                       f' --subgrid={subgrid_fp}'
                                                                       f' --local_res_file={local_res_file}'
                                                                       f' &>> {logfile}',  # redirect and append stdout
                                                                                           # and stderr to logfile
                                                                       get_pty=True)       # execute in pseudoterminal
                    except paramiko.ssh_exception.SSHException as e:
                        # SSH connection has been lost
                        # (remote machine shut down, ssh connection has been killed manually, ...)
                        print(f'[T]\'{thread_name}\': ERROR: {e}')
                        working_grid.at[param_idx, "status"] = "unsolved"
                        connection_lost = True

            # Lock released, thread switching enabled

            # Try to reconnect to host if connection has been lost
            ######################################################
            if connection_lost and connection_lost_counter < 2:
                # Attempt to reconnect while there are still parameter chunks to fetch
                while True:
                    if working_grid.loc[working_grid["status"] == "unsolved"].empty:
                        # Stop thread execution if no more parameters are available
                        return
                    else:
                        print('Attempting to reconnect')
                        pm_client = self.ssh_connect(thread_name, username=getpass.getuser(), print_status=False)
                        if pm_client:
                            print(f'[T]\'{thread_name}\': Reconnected!')
                            connection_lost = False
                            # Escape reconnection loop
                            break
                        t.sleep(30)
                # Jump to the beginning of scheduler loop if reconnection was successful
                continue

            # Wait for remote computation exit status
            # (independent of success or failure)
            #########################################
            exit_status = stdout.channel.recv_exit_status()

            # Update grid status
            ####################
            with self.lock:
                # Remote machine executed worker script successfully
                if exit_status == 0:
                    print(f'[T]\'{thread_name}\': Remote computation finished. Elapsed time: {t.time()-t0:.3f} seconds')
                    print(f'[T]\'{thread_name}\': Updating grid status')
                    try:
                        # Check if data has been written to result file
                        with pd.HDFStore(local_res_file, "r") as store:
                            # Check if temporary result file (hdf5) contains keys
                            if len(store.keys()) == 0:
                                raise KeyError
                            # Check if each key contains data
                            for key in store.keys():
                                if len(store[key].index) == 0:
                                    raise KeyError
                        working_grid.at[param_idx, 'status'] = 'done'
                    except (KeyError, FileNotFoundError):
                        for row in param_idx:
                            if working_grid.loc[row]["err_count"] < 2:
                                working_grid.at[row, "status"] = "unsolved"
                                working_grid.at[row, "err_count"] = working_grid.iloc[row]["err_count"] + 1
                            else:
                                working_grid.at[row, "status"] = "failed"

                # Remote execution was interrupted
                else:
                    print(f'[T]\'{thread_name}\': Remote computation failed with exit status {exit_status}')
                    connection_lost_counter += 1
                    for row in param_idx:
                        if working_grid.loc[row]["err_count"] < 4:
                            working_grid.at[row, "status"] = "unsolved"
                            working_grid.at[row, "err_count"] = working_grid.iloc[row]["err_count"] + 1
                        else:
                            working_grid.at[row, "status"] = "failed"
                    if connection_lost_counter >= 2:
                        print('Excluding worker from pool')
                        return

            # Lock released, thread switching enabled
        # End of scheduler loop

        # TODO: Close pty on remote machine?

    # End of Thread master

    def _fetch_index(self, remaining_params, worker_chunk_size, working_grid):
        param_idx = []
        chunk_count = 0
        chunk_idx = self.chunk_idx
        for i, row in remaining_params.iterrows():
            if chunk_count == worker_chunk_size:
                self.chunk_idx += 1
                return pd.Index(param_idx), chunk_idx
            elif row['chunk_idx'] == -1:
                # Chunk parameter grid
                working_grid.loc[i, 'chunk_idx'] = self.chunk_idx
                param_idx.append(i)
                chunk_count += 1
            else:
                pass
        # Whole parameter grid is chunked
        if len(param_idx) == 0:
            # get all parameters with the same valid chunk index
            chunk_idx = remaining_params.iloc[0]["chunk_idx"]
            param_idx = remaining_params.loc[remaining_params["chunk_idx"] == chunk_idx].index

        return pd.Index(param_idx), chunk_idx

    #################################
    # CGS internal helper functions #
    #################################
    @staticmethod
    def chunk_grid(grid: pd.DataFrame, chunk_size: int):
        """Create a pd.DataFrame from the params dict and assign a chunk index to each parameter combination

        Parameters
        ----------
        grid
            pd.DataFrame with all parameter combinations that can be traversed linearly
        chunk_size
            Number of parameter combinations to be computed simultaneously on one worker

        Returns
        -------
            pd.DataFrame

        """
        # Add status columns to grid
        chunked_grid = grid.copy()
        chunked_grid["status"] = 'unsolved'
        chunked_grid["chunk_idx"] = -1
        chunked_grid["err_count"] = 0

        # Assign a chunk index to each parameter combination
        chunk_count = 0
        while True:
            param_idx = chunked_grid.loc[chunked_grid["status"] == "unsolved"].index[:chunk_size]
            if param_idx.empty:
                break
            else:
                chunked_grid.at[param_idx, 'chunk_idx'] = chunk_count
                chunked_grid.at[param_idx, 'status'] = "chunked"
                chunk_count += 1
        chunked_grid["status"] = 'unsolved'

        return chunked_grid

    @staticmethod
    def check_key_consistency(param_grid: dict, param_map: dict):
        """Check if keys in param_grid and param_map match

        Parameters
        ----------
        param_grid
        param_map

        Returns
        -------
        bool

        """
        grid_key_lst = list(param_grid.keys())
        map_key_lst = list(param_map.keys())
        # Not all keys of parameter map can be found in parameter grid
        if not all((grid_key in map_key_lst for grid_key in grid_key_lst)):
            print("Not all parameter grid keys found in parameter map")
            print("Parameter map keys:")
            print(*list(param_map.keys()))
            print("Parameter grid keys:")
            print(*list(param_grid.keys()))
            return False
        # Parameter grid and parameter map match
        else:
            print("All parameter grid keys found in parameter map")
            return True

    @staticmethod
    def create_cgs_config(fp: str, circuit_template: str, param_map: dict, dt: float, simulation_time: float,
                          inputs: dict, outputs: dict, sampling_step_size: float, add_kwargs: dict):
        """Creates a configfile.json containing a dictionary with all input parameters as key-value pairs

        Parameters
        ----------
        fp
        circuit_template
        param_map
        dt
        simulation_time
        inputs
        outputs
        sampling_step_size
        add_kwargs

        Returns
        -------

        """
        config_dict = {
            "circuit_template": circuit_template,
            "param_map": param_map,
            "dt": dt,
            "simulation_time": simulation_time,
            "outputs": outputs,
            "sampling_step_size": sampling_step_size,
        }
        if inputs:
            config_dict["inputs"] = {str(*inputs.keys()): list(*inputs.values())}
        for key, value in add_kwargs.items():
            config_dict[key] = value
        with open(fp, "w") as f:
            json.dump(config_dict, f, indent=2)

    @staticmethod
    def add_template_information(yaml_fp, hdf5_file):
        """Add opearator information of the circuit template to the global result file"""
        import yaml

        with open(yaml_fp, 'r') as stream:
            try:
                yaml_dict = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)

        for operator_key, operator_value in yaml_dict.items():
            if "Op" in operator_key:
                for temp_key, temp_value in operator_value.items():
                    if isinstance(temp_value, str):
                        hdf5_file.create_dataset(f'/TemplateInfo/{operator_key}/{temp_key}', data=temp_value)
                    if isinstance(temp_value, list):
                        for idx, eq in enumerate(temp_value):
                            hdf5_file.create_dataset(f'/TemplateInfo/{operator_key}/{temp_key}/eq_{idx}', data=eq)
                    elif isinstance(temp_value, dict):
                        for key, value in temp_value.items():
                            try:
                                hdf5_file.create_dataset(f'/TemplateInfo/{operator_key}/{temp_key}/{key}', data=value["default"])
                            except:
                                hdf5_file.create_dataset(f'/TemplateInfo/{operator_key}/{temp_key}/{key}', data=value)

#####################
# Utility functions #
#####################


def linearize_grid(grid: dict, permute: bool = False) -> pd.DataFrame:
    """Turns the grid into a grid that can be traversed linearly, i.e. pairwise.

    Parameters
    ----------
    grid
        Parameter grid.
    permute
        If true, all combinations of the parameter values in grid will be created.

    Returns
    -------
    pd.DataFrame
        Resulting linear grid in form of a data frame.

    """

    arg_lengths = [len(arg) for arg in grid.values()]

    if len(list(set(arg_lengths))) == 1 and not permute:
        return pd.DataFrame(grid)
    else:
        vals, keys = [], []
        for key, val in grid.items():
            vals.append(val)
            keys.append(key)
        new_grid = np.stack(np.meshgrid(*tuple(vals)), -1).reshape(-1, len(grid))
        return pd.DataFrame(new_grid, columns=keys)


def adapt_circuit(circuit: CircuitIR, params: dict, param_map: dict) -> CircuitIR:
    """Changes the parametrization of a circuit.

    Parameters
    ----------
    circuit
        Circuit instance.
    params
        Key-value pairs of the parameters that should be changed.
    param_map
        Map between the keys in params and the circuit variables.

    Returns
    -------
    CircuitIR
        Updated circuit instance.

    """

    for key in params.keys():

        val = params[key]

        for var in param_map[key]['vars']:

            # change variable values on nodes
            nodes = param_map[key]['nodes'] if 'nodes' in param_map[key] else []
            for node in nodes:
                if "/" in var:
                    op, var_name = var.split("/")
                    if op in circuit.nodes[node]['node'].op_graph.nodes:
                        circuit.nodes[node]['node'].op_graph.nodes[op]['variables'][var_name]['value'] = float(val)
                else:
                    for op in circuit.nodes[node]['node'].op_graph.nodes:
                        try:
                            circuit.nodes[node]['node'].op_graph.nodes[op]['variables'][var]['value'] = float(val)
                        except KeyError:
                            pass

            # change variable values on edges
            edges = param_map[key]['edges'] if 'edges' in param_map[key] else []
            if edges and len(edges[0]) < 3:
                for source, target in edges:
                    if var in circuit.edges[source, target, 0]:
                        circuit.edges[source, target, 0][var] = float(val)
            else:
                for source, target, edge in edges:
                    if var in circuit.edges[source, target, edge]:
                        circuit.edges[source, target, edge][var] = float(val)

    return circuit


class StreamTee(object):
    """Tee all stdout to a specified logfile"""
    def __init__(self, stream1, stream2fp):
        stream2 = open(stream2fp, "a")
        self.stream1 = stream1
        self.stream2 = stream2
        self.__missing_method_name = None  # Hack!

    def __getattribute__(self, name):
        return object.__getattribute__(self, name)

    def __getattr__(self, name):
        self.__missing_method_name = name  # Could also be a property
        return getattr(self, '__methodmissing__')

    def __methodmissing__(self, *args, **kwargs):
        # Emit method call to the log copy
        callable2 = getattr(self.stream2, self.__missing_method_name)
        callable2(*args, **kwargs)

        # Emit method call to stdout (stream 1)
        callable1 = getattr(self.stream1, self.__missing_method_name)
        return callable1(*args, **kwargs)



from setuptools import setup, find_packages
from pyrates import __version__

PYRATES_TEAM = "Richard Gast, Daniel Rose"

INSTALL_REQUIREMENTS = ['numpy',
                        'networkx',
                        'pandas',
                        'pyparsing',
                        'ruamel.yaml',
                        ]

CLASSIFIERS = ["Programming Language :: Python :: 3",
               "Programming Language :: Python :: 3.6",
               "Programming Language :: Python :: 3.7",
               "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
               "Operating System :: OS Independent",
               "Development Status :: 3 - Alpha",
               "Intended Audience :: Developers",
               "Intended Audience :: Science/Research",
               "Topic :: Scientific/Engineering",
               ]

with open("README.md", "r", encoding="utf8") as fh:
    DESCRIPTION = fh.read()

setup(name='pyrates',
      version=__version__,
      description='Neural Network Modeling Framework',
      long_description=DESCRIPTION,
      author=PYRATES_TEAM,
      author_email='rgast@cbs.mpg.de',
      license='GPL v3',
      packages=find_packages(),
      zip_safe=False,
      python_requires='>=3.6',
      install_requires=INSTALL_REQUIREMENTS,
      classifiers=CLASSIFIERS,
      include_package_data=True  # include additional non-python files specified in MANIFEST.in
      )



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network models and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

# system imports
import os
import sys
import ast
import json
import time
import argparse

# external imports
import numpy as np
import pandas as pd
from pyrates.utility.grid_search import grid_search


def main(_):
    # disable TF-gpu warnings
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    t_total = time.time()

    # Load command line arguments and create logfile
    ################################################
    print("")
    print("***LOADING COMMAND LINE ARGUMENTS***")
    t0 = time.time()

    config_file = FLAGS.config_file
    subgrid = FLAGS.subgrid
    local_res_file = FLAGS.local_res_file

    print(f'Elapsed time: {time.time()-t0:.3f} seconds')

    # Load grid search configuration parameters from config file
    ############################################################
    print("")
    print("***LOADING GLOBAL CONFIG FILE***")
    t0 = time.time()

    with open(config_file) as g_conf:
        global_config_dict = json.load(g_conf)

        circuit_template = global_config_dict['circuit_template']
        param_map = global_config_dict['param_map']
        sampling_step_size = global_config_dict['sampling_step_size']
        dt = global_config_dict['dt']
        simulation_time = global_config_dict['simulation_time']
        # 'inputs' and 'outputs' are optional parameters
        try:
            inputs_temp = global_config_dict['inputs']
            if inputs_temp:
                inputs = {}
                for key, value in inputs_temp.items():
                    inputs[ast.literal_eval(key)] = list(value)
            else:
                inputs = {}
        except KeyError:
            inputs = {}
        try:
            outputs_temp = global_config_dict['outputs']
            if outputs_temp:
                outputs = {}
                for key, value in outputs_temp.items():
                    outputs[str(key)] = tuple(value)
            else:
                outputs = {}
        except KeyError:
            outputs = {}

    print(f'Elapsed time: {time.time()-t0:.3f} seconds')

    # Load parameter subgrid from subgrid file
    ##########################################
    print("")
    print("***PREPARING PARAMETER GRID***")
    t0 = time.time()

    param_grid = pd.read_hdf(subgrid, key="subgrid")

    # grid_search() can't handle additional columns in the parameter grid
    param_grid = param_grid.drop(['status', 'chunk_idx', 'err_count'], axis=1)
    print(f'Elapsed time: {time.time()-t0:.3f} seconds')

    # Compute parameter subgrid using grid_search
    #############################################
    print("")
    print("***COMPUTING PARAMETER GRID***")
    t0 = time.time()

    results = grid_search(circuit_template=circuit_template,
                          param_grid=param_grid,
                          param_map=param_map,
                          inputs=inputs,
                          outputs=outputs,
                          sampling_step_size=sampling_step_size,
                          dt=dt,
                          simulation_time=simulation_time)

    out_vars = results.columns.levels[-1]

    print(f'Total parameter grid computation time: {time.time()-t0:.3f} seconds')

    # Post process results and write data to local result file
    ##########################################################
    print("")
    print("***POSTPROCESSING AND CREATING RESULT FILES***")
    t0 = time.time()

    with pd.HDFStore(local_res_file, "w") as store:
        for out_var in out_vars:
            key = out_var.replace(".", "")
            res_lst = []

            # Order results according to rows in parameter grid
            ###################################################
            for i, idx in enumerate(param_grid.index):
                idx_list = param_grid.iloc[i].values.tolist()
                idx_list.append(out_var)

                result = results.loc[:, tuple(idx_list)].to_frame()
                result.columns.names = results.columns.names
                res_lst.append(result)

            result_ordered = pd.concat(res_lst, axis=1)

            # Postprocess ordered results (optional)
            ########################################

            # Write DataFrames to local result file
            ######################################
            store.put(key=key, value=result_ordered)

    # TODO: Copy local result file back to master if needed

    print(f'Result files created. Elapsed time: {time.time()-t0:.3f} seconds')
    print("")
    print(f'Total elapsed time: {time.time()-t_total:.3f} seconds')


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--config_file",
        type=str,
        default="~/worker_test/config_file.json",
        help="Config file with all necessary data to start grid_search() except for parameter grid"
    )

    parser.add_argument(
        "--subgrid",
        type=str,
        default="~/worker_test/subgrid.h5",
        help="Path to csv-file with sub grid to compute on the remote machine"
    )

    parser.add_argument(
        "--local_res_file",
        type=str,
        default="~/worker_test/result.h5",
        help="hdf5-file to save results to"
    )

    FLAGS = parser.parse_args()

    main(sys.argv)


def postprocessing_1(data_):
    """Compute spike frequency based on frequency in PSD with the highest energy"""

    from pyrates.utility import plot_psd
    import matplotlib.pyplot as plt

    # Store columns for later reconstruction
    cols = data_.columns

    # Plot_psd expects only the out_var as column value
    index_dummy = pd.Index([cols.values[-1][-1]])
    data_.columns = index_dummy
    if not any(np.isnan(data_.values)):
        _ = plot_psd(data_, tmin=1.0, show=False)
        pow_ = plt.gca().get_lines()[-1].get_ydata()
        freqs = plt.gca().get_lines()[-1].get_xdata()
        plt.close()
        max_freq = freqs[np.argmax(pow_)]
        freq_pow = np.max(pow_)
        temp = [max_freq, freq_pow]
        psd = pd.DataFrame(temp, index=['max_freq', 'freq_pow'], columns=cols)
        data_.columns = cols
        return psd

    data_.columns = cols

    # Return empty DataFrame if data contains NaN
    return pd.DataFrame(columns=cols)


def postprocessing_2(data, simulation_time):
    """Compute spike frequency based on average number of spikes (local maxima) per second"""

    import scipy.signal as sp

    cols = data.columns

    np_data = np.array(data.values)

    peaks = sp.argrelextrema(np_data, np.greater)
    num_peaks_temp = int(len(peaks[0]) / simulation_time)

    return pd.DataFrame(num_peaks_temp, index=['num_peaks'], columns=cols)


def postprocessing_3(data, dt):
    """Compute spike frequency based on average time between spikes (local maxima)"""

    import scipy.signal as sp

    cols = data.columns

    np_data = np.array(data.values)

    peaks = sp.argrelextrema(np_data, np.greater)
    diff = np.diff(peaks[0])
    diff = np.mean(diff) * dt
    return 1 / diff




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
from typing import Iterator, Dict
from networkx import DiGraph, find_cycle, NetworkXNoCycle

from pyrates import PyRatesException

__author__ = "Daniel Rose"
__status__ = "Development"


class OperatorGraph(DiGraph):
    """Intermediate representation for nodes and edges."""

    def __init__(self, operators: dict = None, template: str = ""):

        super().__init__()
        if not operators:
            operators = {}
        all_outputs = {}  # type: Dict[str, dict]
        # op_inputs, op_outputs = set(), set()

        for key, item in operators.items():

            op_instance = item["operator"]
            op_variables = item["variables"]

            # add operator as node to local operator_graph
            self.add_node(key, operator=op_instance, variables=op_variables)

            # collect all output variables
            out_var = op_instance.output

            # check, if variable name exists in outputs and create empty list if it doesn't
            if out_var not in all_outputs:
                all_outputs[out_var] = {}

            all_outputs[out_var][key] = out_var
            # this assumes input and output variables map on each other by equal name
            # with additional information, non-equal names could also be mapped here

        # link outputs to inputs
        for op_key, op in self.operators(get_ops=True):
            for in_var in op.inputs:
                if in_var in all_outputs:
                    # link all collected outputs of given variable in inputs field of operator
                    for predecessor, out_var in all_outputs[in_var].items():
                        # add predecessor output as source; this would also work for non-equal variable names
                        if predecessor not in op.inputs[in_var]["sources"]:
                            op.inputs[in_var]["sources"].append(predecessor)
                        self.add_edge(predecessor, op_key)
                else:
                    pass  # means, that 'source' will remain an empty list and no incoming edge will be added

        try:
            find_cycle(self)
        except NetworkXNoCycle:
            pass
        else:
            raise PyRatesException("Found cyclic operator graph. Cycles are not allowed for operators within one node "
                                   "or edge.")

    def getitem_from_iterator(self, key: str, key_iter: Iterator[str]):
        """
        Helper function for Python magic __getitem__. Accepts an iterator that yields string keys. If `key_iter`
        contains one key, an operator will be (looked for and) returned. If it instead contains two keys, properties of
        a variable that belong to an operator is returned.

        Parameters
        ----------
        key
        key_iter

        Returns
        -------
        item
            operator or variable properties
        """

        try:
            var = next(key_iter)
        except StopIteration:
            # no variable specified, so we return an operator
            item = self.nodes[key]["operator"]
        else:
            # variable specified, so we return variable properties instead
            item = self.nodes[key]["variables"][var]

        return item

    def __iter__(self):
        """Return an iterator containing all operator labels in the operator graph."""
        return iter(self.nodes)

    def operators(self, get_ops=False, get_vars=False):
        """Alias for self.nodes"""

        if get_ops and get_vars:
            return ((key, data["operator"], data["variables"]) for key, data in self.nodes(data=True))
        elif get_ops:
            return ((key, data["operator"]) for key, data in self.nodes(data=True))
        elif get_vars:
            return ((key, data["variables"]) for key, data in self.nodes(data=True))
        else:
            return self.nodes
